from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo
import re


class SpecTypeDeJager(object):
    r"""
    Spectral type calibration from de Jager and Nieuwenhuijzen.

    This class implements the spectral type calibration presented by
    de Jager and Nieuwenhuijzen 1987, A&A 177, 217-227 (DJ87). Specifically,
    the authors calibrate the relation between spectral type and stellar
    luminosity and effective temperature for a wide range of spectral types
    and luminosity classes.

    DJ87 give their results in the form of a polynomial relation with 20
    coefficients (Eqs. 2a and 2b) as well as in the form of their
    Tables 5 and 6.

    .. Note:: Tables 5 and 6 in DJ87 are calculated on the basis of a
              40 parameter solution, which is not given in the paper.
              The numbers based on the 20 parameter polynomial solution
              deviate from the tabulated numbers by typically a few percent
              for the effective temperature and 10 percent for the
              luminosities (not their logarithm).

    """

    def __init__(self):
        self._typeLetter = ["O", "B", "A", "F", "G", "K", "M"]
        # Generate all valid types (without LK)
        self._types = [l+str(n)
                       for l in self._typeLetter for n in list(smo.range(10))]
        # Adjust to dJ87
        self._types.pop(0)
        self._types.append("M10")

        # Table to resolve variable s (Table 1)
        self._stable = [("O1", 0.1, 0.1),
                        ("O9", 0.9, 0.3),
                        ("B2", 1.8, 0.15),
                        ("A0", 3.0, 0.1),
                        ("F0", 4.0, 0.1),
                        ("G0", 5.0, 0.05),
                        ("K0", 5.5, 0.1),
                        ("M0", 6.5, 0.2)]
        # Table to resolve b (Table 2)
        self._btable = {"V": 5., "IV": 4., "III": 3., "II": 2.,
                        "Ib": 1.4, "Iab": 1.0, "Ia": 0.6, "Ia+": 0.0}
        # Table to resolve Chebychev polynomials (use self._cij[i,j])
        self._cij = np.array([[3.82573, -2.13868, -0.46357,  0.02076, -0.11937],
                              [-1.55607, -1.89216, -0.96916, -0.08869, -0.20423],
                              [1.05165,  0.42330, -0.94379, -0.07438, np.nan],
                              [-0.01663, -0.20024, -0.18552, np.nan,   np.nan],
                              [-0.07576, -0.10934, np.nan, np.nan,   np.nan],
                              [0.11008, np.nan, np.nan, np.nan,   np.nan]])
        # Table to resolve Chebychev polynomials (use self._dij[i,j])
        self._dij = np.array([[3.96105,  0.03165, -0.02963,  0.01307, -0.01172],
                              [-0.62945,  0.02596, -0.06009,  0.01881, -0.01121],
                              [0.14370, -0.00977, -0.03265,  0.01649, np.nan],
                              [0.00791,  0.00076, -0.03006, np.nan,   np.nan],
                              [0.00723, -0.02621, np.nan, np.nan,   np.nan],
                              [0.02755, np.nan, np.nan, np.nan,   np.nan]])

    def _indexSpecType(self, spt):
        r"""
        Get index of spectral type in list and check validity.

        Returns
        -------
        index : int
            Position in self._types list
        """
        if not spt in self._types:
            raise (PE.PyAValError("'" + str(spt) + "' is not a valid spectral type.",
                                  where="SpecTypeDeJager",
                                  solution="Use either of: " + ', '.join(self._types)))
        return self._types.index(spt)

    def _resolve_s(self, bt, st):
        r"""
        Resolve numerical value of variable s.

        Parameters
        ----------
        bt : string
            Spectral type (e.g., F3), no LK included
        st : float
            Subtype (e.g., 3.5 from F3.5)

        Returns
        -------
        s : float
            Value of s
        """
        spi = self._indexSpecType(bt)
        for i in smo.range(len(self._stable)-1, -1, -1):
            # Range start index
            rsi = self._indexSpecType(self._stable[i][0])
            if spi >= rsi:
                break
        s = self._stable[i][1] + float(spi+(st-int(st))-rsi)*self._stable[i][2]
        return s

    def _resolve_b(self, lk):
        r"""
        Resolve numerical value of variable b.

        Parameters
        ----------
        lk : string
            Luminosity class (e.g., IV)

        Returns
        -------
        b : float
            The value of b
        """
        if not lk in self._btable:
            raise (PE.PyAValError("'" + str(lk) + "' is not a valid luminosity class.",
                                  where="SpecTypeDeJager",
                                  solution="Use either of: " + ', '.join(list(self._btable))))
        return self._btable[lk]

    def _cheby(self, i, x):
        r"""
        Evaluate the Chebychev polynomial.
        """
        return np.cos(i*np.arccos(x))

    def _decomposeSPT(self, spt):
        r"""
        Decomposes spectral type into 'basic type' and subtype.

        Parameters
        ----------
        spt : string
            Spectral type (e.g, F8.5)

        Returns
        -------
        Basic type : string
            Type with integer subtype (e.g., F8)
        Subtype : float
            Subtype as a float (e.g., 8.5)
        """
        r = re.match(r"^\s*([OBAFGKM])(([\d])(\.\d+)?)\s*$", spt)
        if r is None:
            raise (PE.PyAValError("'" + str(spt) + "' is not a valid spectral type.",
                                  where="SpecTypeDeJager",
                                  solution="Use either of (also with float subtype): " + ', '.join(list(self._btable))))
        return r.group(1)+r.group(3), float(r.group(2))

    def lumAndTeff(self, spt, lk):
        r"""
        Get effective temperature and luminosity.

        Parameters
        ----------
        spt : string
            The spectral type (may include float subtype).
            For instance, F3, F3.5, or K2
        lk : string, {V, IV, III, II, Ib, Iab, Ia, Ia+}
            The luminosity class.

        Returns
        -------
        log10(L) : float
            The base-10 logarithm of the luminosity in units
            of the solar luminosity.
        log10(teff) : float
            The base-10 logarithm of the effective temperature.
        """
        bt, st = self._decomposeSPT(spt)
        s = self._resolve_s(bt, st)
        b = self._resolve_b(lk)
        # Normalize to -1,+1 (Eqs. 2ab in dJ87)
        snorm = (s-4.25)/4.25
        bnorm = (b-2.5)/2.5

        # Prepare result
        llum = 0.0
        lteff = 0.0

        for n in smo.range(0, 6):
            for i in smo.range(0, n+1):
                j = n - i
                t = self._cheby(i, snorm)*self._cheby(j, bnorm)
                try:
                    llum += (self._cij[i, j] * t)
                    lteff += (self._dij[i, j] * t)
                except:
                    pass
        return llum, lteff



from __future__ import print_function, division
import unittest


class TestExampleSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_pyaPickerExample(self):
        from PyAstronomy import pyaGui
        import numpy as np

        # Data for the plot
        x = np.linspace(0.0, 10.0, 100)
        y = np.exp(-x / 10.0)

        # Create Picker instance
        pp = pyaGui.Picker()

    #     # Create the plot in which points
    #     # are to be selected
    #     pp.a.plot(x, y, 'b.-')
    #
    #     points = pp.pick()
    #
    #     print("Number of selected points: ", len(points))
    #     print("Selected points (x, y):")
    #     for p in points:
    #       print("%g, %g" % p)

    def testsanity_modelExplorer(self):
        """
        Check sanity of the model explorer example
        """
        import numpy as np
        from PyAstronomy import pyaGui
        from PyAstronomy import funcFit as fuf

        # Create a Gaussian fitting instance
        # and set some parameters
        gg = fuf.GaussFit1d()
        gg["A"] = 1.0
        gg["sig"] = 0.5

        # Let A and mu be free during a fit
        gg.thaw(["A", "mu"])

        # Create some artificial data
        x = np.linspace(-2.0, 2.0, 100)
        yerr = np.ones(len(x)) * 0.01
        y = gg.evaluate(x) + np.random.normal(0.0, 0.01, len(x))

        # In order to use the interactive explorer, you
        # need a class, which plots the model. The default
        # class for this purpose is "FFModelPlotFit", which
        # needs the x and y values. Optionally, you can specify
        # errors via `yerr`. Depending on the setting for
        # "withResiduals", the residuals will be shown or not.
        mp = pyaGui.FFModelPlotFit(x, y, yerr=yerr, withResiduals=True)

        # Use the function ffmodelExplorer (note the lowercase letters)
        # to create an instance of the FFModelExplorer class, which
        # needs to be given the model (gg in this case) and
        # the plotter (and fitter), which we created above.
        #
        g = pyaGui.ffmodelExplorer(gg, mp)

    #    g.show()

    #  def testsanity_modelExplorerCustomPlot(self):
    #    """
    #      Check custom plot example of model explorer.
    #    """
    #    import numpy as np
    #    from PyAstronomy import pyaGui
    #    from PyAstronomy import funcFit as fuf
    #
    #
    #    class CustomPlotFit:
    #      """
    #        A custom class for plotting and fitting.
    #
    #        This class uses several panels
    #        can be used for plotting.
    #
    #        Parameters
    #        ----------
    #        x, y : arrays
    #            The data.
    #        n : int
    #            The number of panels.
    #      """
    #
    #      def __init__(self, x, y, n):
    #        self.x = x
    #        self.y = y
    #        self.n = n
    #        self.a = None
    #        self.color = {0:"b", 1:"r", 2:"g"}
    #
    #      def plot(self, f, odf):
    #        """
    #          Plot the model into n panels.
    #
    #          Parameters
    #          ----------
    #          f : Matplotlib Figure
    #              Figure on which to plot.
    #          odf : Model class instance
    #              The model to be plotted.
    #        """
    #
    #        if self.a is None:
    #          # Create axis instances for plotting
    #          self.a = []
    #          for i in range(self.n):
    #            self.a.append(f.add_subplot(self.n, 1, i+1))
    #
    #        model = odf.evaluate(self.x)
    #        for i in range(self.n):
    #          self.a[i].cla()
    #          self.a[i].plot(self.x, self.y, self.color[i % 3]+'p')
    #          self.a[i].plot(self.x, model, 'k--')
    #
    #      def fit(self, odf):
    #        """
    #          Fit the model.
    #
    #          Parameters
    #          ----------
    #          odf : Model class instance
    #              The model to be fitted.
    #        """
    #        odf.fit(self.x, self.y)
    #
    #
    #    # Create a Gaussian fitting instance
    #    # and set some parameters
    #    gg = fuf.GaussFit1d()
    #    gg["A"] = 1.0
    #    gg["sig"] = 0.5
    #
    #    # Let A and mu be free during a fit
    #    gg.thaw(["A", "mu"])
    #
    #    # Create some artificial data
    #    x = np.linspace(-2., 2., 100)
    #    yerr = np.ones(len(x))*0.01
    #    y = gg.evaluate(x) + np.random.normal(0.,0.01, len(x))
    #
    #    # Use the custom class for plotting and fitting, which
    #    # will show the plot in 3 panels.
    #    cm = CustomPlotFit(x, y, 3)
    #
    #    # Create an instance of the FFModelExplorer class, which
    #    # needs to be given the model (gg in this case) and
    #    # the plotter (and fitter), which we created above.
    #    g = pyaGui.FFModelExplorer(gg, cm)
    ##     g.show()

    def testsanity_InteractiveContinuum(self):
        """
        Checking sanity of interactive continuum.
        """
        from PyAstronomy import pyaGui as pg
        import numpy as np
        import matplotlib.pylab as plt

        # Create some artificial data/spectrum
        x = np.arange(1000)
        y = np.random.random(1000) + 1000.0 + np.sin(x / 80.0)
        y += np.random.random(len(x))

        # Construct an instance of the interactive normalization GUI
        cf = pg.ContinuumInteractive(x, y)

        # The 'plot' command can be used to draw something on the canvas
        # on which normalization points are selected. Uncomment the following
        # line to see an example.
        #
        # cf.plot([4,400,600], [1001,1000,999], 'k--')

        # The 'plotNorm' command can be used to draw something on the
        # canvas showing the normalized spectrum (if display is enabled).
        # Uncomment the following line to see an example.
        #
        # cf.plotNorm([0.,1000.], [1.001,0.999], 'g:')

        # Opens the GUI and starts the interactive session.

    #    c = cf.findContinuum()

    # The outcome is a dictionary with the following keys:
    #
    #            - points: A list of two-float tuples holding the
    #                      x,y location of the selected points.
    #            - continuum : Array holding the continuum estimate
    #                          at the given x-values.
    #            - splineKind : A string specifying the selected
    #                           spline option.
    #            - normalizedData : An array holding the normalized
    #                               data.
    #    plt.title("Normalized data")
    #    plt.plot(x, c["normalizedData"], 'b.--')
    #    plt.show()

    def testsanity_IAGVFitExample(self):
        """
        Checking example of IAGVFit
        """
        from PyAstronomy import pyaGui
        from PyAstronomy import funcFit as fuf
        import numpy as np

        # Data for the plot
        x = np.linspace(5000.0, 5010, 200)
        y = np.ones(len(x))
        yerr = np.ones(len(x)) * 0.01
        y += np.random.normal(0.0, 0.01, len(x))

        gf = fuf.GaussFit1d()
        gf["A"] = -0.3
        gf["mu"] = 5004.0
        gf["sig"] = 0.2
        y += gf.evaluate(x)

        # Create interactive fitter
        igv = pyaGui.IAGVFit(x, y, yerr=yerr, mode="gauss")


#     r = igv.interactiveFit()

#     print("Parameters of the fit: ", r[0])
#     print("Parameters of active component: ", r[1])
#     print("No. of components: ", r[2])



from __future__ import print_function, division
import unittest
import os


class TestModSuiteSanity(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        # Clean up example output from KeplerEllipseModel example
        if os.path.isfile("kemExample.emcee"):
            os.remove("kemExample.emcee")

    def testsanity_rmcl_model(self):
        """ Checking sanity of RmcL calculation (example) """
        # Import some unrelated modules
        from numpy import arange, pi
        import matplotlib.pylab as plt
        # ... and the model suite
        from PyAstronomy import modelSuite as ms

        # Create Rossiter-McLaughlin object
        rmcl = ms.RmcL()
        # Set parameters
        rmcl.assignValue({"a": 6.7, "lambda": 7.2 / 180.0 * pi, "epsilon": 0.5,
                          "P": 1.74, "T0": 0.2, "i": 87.8 / 180. * pi,
                          "Is": 90.0 / 180.0 * pi, "Omega": 1.609e-5, "gamma": 0.2})
        # Choose some time axis and calculate model
        time = arange(100) / 100.0 * 0.2 + 0.1
        rv = rmcl.evaluate(time)

        # Let's see what happened...
        plt.ylabel("Radial velocity [stellar-radii/s]")
        plt.xlabel("Time [d]")
        plt.plot(time, rv, '.')
#     plt.show()

    def testsanity_rmcl_fit(self):
        """ Checking sanity of RmcL fit (example) """
        # Import some unrelated modules
        from numpy import arange, pi, random
        import matplotlib.pylab as plt
        # ... and the model suite
        from PyAstronomy import modelSuite as ms

        # Create Rossiter-McLaughlin object
        rmcl = ms.RmcL()
        # Set parameters
        rmcl.assignValue({"a": 6.7, "lambda": 7.2 / 180.0 * pi, "epsilon": 0.5,
                          "P": 1.74, "T0": 0.2, "i": 87.8 / 180. * pi,
                          "Is": 90.0 / 180.0 * pi, "Omega": 1.609e-5, "gamma": 0.2})
        # Choose some time axis and calculate model
        time = arange(100) / 100.0 * 0.2 + 0.1
        rv = rmcl.evaluate(time)

        # Add some noise.
        rv += random.normal(0.0, 0.05 * rv.max(), rv.size)

        # Assign guess parameters
        rmcl.assignValue({"a": 6.0, "lambda": 7.2 / 180.0 * pi, "epsilon": 0.5,
                          "P": 1.74, "T0": 0.17, "i": 87.8 / 180. * pi,
                          "Is": 90.0 / 180.0 * pi, "Omega": 1.609e-5, "gamma": 0.2})

        # Thaw parameters and fit
        rmcl.thaw(["a", "T0"])
        rmcl.fit(time, rv)

        # Investigate the outcome
        rmcl.parameterSummary()

        # Let's see what happened...
        plt.ylabel("Radial velocity [stellar-radii/s]")
        plt.xlabel("Time [d]")
        plt.plot(time, rv, '.')
        plt.plot(time, rmcl.model, 'r--')
        plt.legend(["Observation", "Model"])
#     plt.show()

    def testsanity_rmclell_calc(self):
        """
        Checking sanity of RmcLell (example)
        """
        # Import some unrelated modules
        from numpy import arange, pi
        import matplotlib.pylab as plt
        # ... and the model suite
        from PyAstronomy import modelSuite as ms
        
        # Create Rossiter-McLaughlin object (circular orbit)
        rmcl = ms.RmcL()
        # and one for an elliptical orbit
        rmel = ms.RmcLell()
        
        # Assign parameter values
        rmcl.assignValue({"a":6.7, "lambda":7.2/180.0*pi, "epsilon":0.5, \
                            "P":1.74, "T0":0.2, "i":87.8/180.*pi, \
                            "Is":90.0/180.0*pi, "Omega":1.609e-5, "gamma":0.2})
        rmel.assignValue({"a":6.7, "lambda":7.2/180.0*pi, "epsilon":0.5, \
                            "P":1.74, "tau":0.2, "i":87.8/180.*pi, "w":-90/180.*pi, \
                            "e":0.05, "Is":90.0/180.0*pi, "Omega":1.609e-5, "gamma":0.2})
        
        
        # Choose some time axis and calculate model
        time = arange(100)/100.0 * 0.2 + 0.1
        rvc = rmcl.evaluate(time)
        rve = rmel.evaluate(time)
        
        # Let's see what happened...
#         plt.ylabel("Radial velocity [stellar-radii/s]")
#         plt.xlabel("Time [d]")
#         plt.plot(time, rvc, 'b.-', label="circular")
#         plt.plot(time, rve, 'r.-', label="elliptical")
#         plt.legend()
#         plt.show()

    def testsanity_rmcl_vs_rmclell(self):
        """ Cross-checking Rmcl and RmcLell """
        from numpy import arange, pi
        import numpy as np
        # ... and the model suite
        from PyAstronomy import modelSuite as ms
        
        # Create Rossiter-McLaughlin object
        rmcl = ms.RmcL()
        r2 = ms.RmcLell()
        
        np.random.seed(9234667)
        
        for i in range(10):
        
            a = np.random.random()*5 + 3
            l = np.random.random()*180 - 90
            inc = np.random.random()*2 + 88
            Omega = 1e-5 + np.random.random()*1e-5
        
            # Set parameters
            rmcl.assignValue({"a":a, "lambda":l/180.0*pi, "epsilon":0.5, \
                            "P":1.74, "T0":0.2, "i":inc/180.*pi, \
                            "Is":80.0/180.0*pi, "Omega":Omega, "gamma":0.2})
        
            # Set parameters
            r2.assignValue({"a":a, "lambda":l/180.0*pi, "epsilon":0.5, \
                            "P":1.74, "tau":0.2, "i":inc/180.*pi, \
                            "Is":80.0/180.0*pi, "Omega":Omega, "gamma":0.2,
                            "e":0.0, "w":-90/180*pi})
        
            # Choose some time axis and calculate model
            time = arange(20)/20.0 * 0.2 - 0.1 + rmcl["T0"]
            rv = rmcl.evaluate(time)
            rv2 = r2.evaluate(time)
            
            d = np.max(np.abs(rv-rv2))
            m = np.max(np.abs(rv))
            self.assertAlmostEqual(d/m, 0.0, delta=1e-8, msg="Elliptical and circular orbit solution for RmcL and RmcLell do not match. " + \
                                   str(r2.parameters()))

    def testsanity_SinRadVel(self):
        # Import some unrelated modules
        from numpy import arange, random, ones
        import matplotlib.pylab as plt
        # ... and now the radVel module
        from PyAstronomy.modelSuite import radVel as rv

        # Create Radial Velocity object
        r = rv.SinRadVel()
        # Set parameters
        r.assignValue({"P": 1.8, "T0": 0.25, "K": 0.5, "rv0": 10.0})
        # Choose some time axis and calculate model
        time = arange(100) / 100.0 * 3.0 - 1.5
        y = r.evaluate(time)

        # Create some faked data by adding noise
        rvData = y + random.normal(0.0, 0.05, y.size)

        # Randomize starting parameters for fit
        for p, v in r.parameters().items():
            r[p] = v + (random.random() - 0.5) * v
        # Show starting values
        print("Starting values for fit:")
        r.parameterSummary()

        # Thaw all parameters
        r.thaw(list(r.parameters().keys()))
        # Start the fit
        r.fit(time, rvData, yerr=ones(y.size) * 0.05)

        # Show fit results
        print("Fitted values:")
        r.parameterSummary()

        # Let's see what happened...
        plt.ylabel("Radial velocity [km/s]")
        plt.xlabel("Radial velocity [d]")
        plt.errorbar(time, rvData, yerr=ones(y.size) * 0.05, fmt='b.')
        plt.plot(time, y, 'r-')
#     plt.show()

    def testsanity_KeplerEllipseModel(self):
        """ Sanity of Kepler Ellipse Model example """
        from PyAstronomy.modelSuite import KeplerEllipseModel
        import numpy as np
        import matplotlib.pylab as plt
        
        # Create a model class instance
        # In this case, we are only interested
        # in the x- and z-components of the orbit
        # solution.
        kem = KeplerEllipseModel(relevantAxes="xz")
        
        # Setting some guess parameters
        kem["a"] = 7.8
        kem["per"] = 12.3
        kem["e"] = 0.07
        kem["tau"] = 0.745
        kem["Omega"] = 143.
        kem["w"] = 0.2
        kem["i"] = 92.0
        
        # Evaluate the model
        time = np.linspace(0, kem["per"], 20)
        model = kem.evaluate(time)
        # Note that the model has twice the number of points
        # compared to the time axis. This is because it contains
        # the data for two axes
        print("Used " + str(len(time)) + " time points")
        print("-> length of model: ", len(model))
        
        # Isolating the model for the x-axis, i.e.,
        # every second data point starting from the
        # beginning.
        xmodel = model[0::2]
        # Isolating the model for the y-axis
        ymodel = model[1::2]
        
        # Use the model to obtain mock data
        # by introducing some scatter
        data = model + np.random.normal(0., 0.5, model.size)
        # Plot the resulting "data"
        plt.title("Kepler Ellipse Model --- Example")
        plt.errorbar(data[0::2], data[1::2], xerr=np.ones(20)*0.5,
                     yerr=np.ones(20)*0.5, fmt="bp")
        
        # Use MCMC to sample from the posterior
        # Specify free parameters
        kem.thaw(["a", "per", "e", "tau", "Omega", "w", "i"])
        # Specify starting values
        X0 = {}
        steps = {}
        for p in kem.freeParameters():
            X0[p] = kem[p]
            steps[p] = kem[p] / 20.
        
        lims = {"a": [5., 10.], "per": [10., 15.], "e": [0., 1.], "tau": [0.5, 1.],
                "Omega": [0., 360.], "w": [-5., 5.], "i": [90., 95.]}
        
        # Generate functions serving as uniform priors with lower and upper limit
        def getprior(ll, ul):
            """ lower (ll) and upper (ul) limit """
            def prior(pardict, p):
                if pardict[p] < ll:
                    return -np.inf
                elif pardict[p] > ul:
                    return -np.inf
                else:
                    return 0
            return prior
        
        # Define priors for parameters according to lims
        priors = {par:getprior(l[0], l[1]) for par, l in lims.items()}
        
        kem.fitEMCEE(time, data, yerr=np.ones(len(data))*0.5, dbfile="kemExample.emcee", priors=priors)
        
        # Plot the lowest deviance model
        ldmodel = kem.evaluate(np.linspace(0, kem["per"], 200))
        plt.plot(ldmodel[0::2], ldmodel[1::2], 'r--')
        
        #plt.show()



#  def testsanity_atanProfile(self):
#    from PyAstronomy import modelSuite as ms
#    import numpy as np
#    import matplotlib.pylab as plt
#
#    # Create an instance of the AtanProfile ...
#    ap = ms.AtanProfile()
#    # ... and define some starting values
#    ap["A"] = 1.0
#    ap["mu"] = 5.0
#    ap["scale"] = 0.4
#    ap["sig"] = 5.0
#
#    # Plot profile on given x-axis
#    x = np.linspace(-5,15,100)
#    plt.plot(x, ap.evaluate(x), 'b.-')
#
#    # Determine the locations of the inflection
#    # points
#    print "Inflection points: ", ap.inflectionPoints()
#
#    # Create instance of damped profile and copy
#    # the values from the first profile
#    apd = ms.AtanProfileDamped()
#    for p, v in ap.parameters().iteritems():
#      apd[p] = v
#
#    # Specify the additional damping parameter
#    apd["tau"] = 2.0
#    # and plot
#    plt.plot(x, apd.evaluate(x), 'r.-')
# plt.show()

    def testsanity_lineListGaussModel(self):
        """
        Checking example of line list Gauss model
        """
        from PyAstronomy import modelSuite as ms
        import numpy as np
        import matplotlib.pylab as plt

        # Create our line list with 4 line
        lineList = np.zeros((4, 3))
        # Assign wavelengths (in A)
        lineList[0, 0] = 5002.37
        lineList[1, 0] = 5005.9
        lineList[2, 0] = 5007.52
        lineList[3, 0] = 5007.64
        # Assign EWs (in A)
        lineList[0, 1] = 0.01
        lineList[1, 1] = 0.05
        lineList[2, 1] = 0.009
        lineList[3, 1] = 0.12
        # Assign depths (0-1)
        lineList[0, 2] = 0.97
        lineList[1, 2] = 0.9
        lineList[2, 2] = 0.99
        lineList[3, 2] = 0.35

        wvl = np.arange(5000., 5010., 0.01)

        # Get an instance of the LLGauss class
        llg = ms.LLGauss(lineList)
        # Have a look at the model parameters
        llg.parameterSummary()
        # Evaluate the model
        m1 = llg.evaluate(wvl)
        # Now apply rotational broadening [km/s]
        # with limb-darkening of 0.6
        llg["vsini"] = 61.0
        llg["eps"] = 0.6
        # and evaluate again
        mvsini = llg.evaluate(wvl)
        # Next, apply a Doppler shift [km/s]
        llg["vrad"] = -32.7
        # and evaluate
        mvrad = llg.evaluate(wvl)
        # Plot the results
        plt.subplot(2, 1, 1)
        plt.plot(wvl, m1, 'b.-')
        plt.plot(wvl, mvsini, 'g.-')
        plt.plot(wvl, mvrad, 'y.-')

        # Now use the model for fitting
        # We need "data" ...
        data = llg.evaluate(wvl)
        # ... with noise
        data += np.random.normal(0.0, 0.01, len(data))
        # Lets modify the strengths of the Gaussians
        # and get it back.
        for i in range(llg.numberOfLines()):
            llg["A" + str(i + 1)] += np.random.normal(0.0, 0.1)
        # Use all line strengths for fitting
        llg.thawLineStrengths()
        # and fit
        llg.fit(wvl, data)
        # Plot the result
        plt.subplot(2, 1, 2)
        plt.errorbar(wvl, data, yerr=np.ones(len(wvl)) * 0.01, fmt='bp')
        plt.plot(wvl, llg.evaluate(wvl), 'r--')
#    plt.show()

    def testsanity_VoigtAstroPExample(self):
        """
        Sanity of VoigtAstroP example
        """
        from PyAstronomy import modelSuite as ms
        import numpy as np
        import matplotlib.pylab as plt

        # Obtain an object of type VoigtAstroP ...
        v = ms.VoigtAstroP()
        # ... and set some parameters
        v["b"] = 87.7
        v["f"] = 0.5
        v["w0"] = 1214.0
        # Damping constant [cm]
        v["gamma"] = 2e-9

        # Generate wavelength axis ...
        wvl = np.linspace(1212., 1216., 200)
        # ... and evaluate model
        m = v.evaluate(wvl)

        # Plot result
        plt.plot(wvl, m, 'b.-')
#    plt.show()


    def testsanity_VoigtAstroP_R_Example(self):
        """
        Sanity of VoigtAstroP example with instrumental resolution
        """

        from PyAstronomy import modelSuite as ms
        import numpy as np
        import matplotlib.pylab as plt
        
        # Obtain an object of type VoigtAstroP ...
        v = ms.VoigtAstroP()
        # ... and set some parameters
        v["b"] = 40.7
        v["f"] = 0.5
        v["w0"] = 1214.0
        # Damping constant [cm]
        v["gamma"] = 2e-9
        
        # Generate wavelength axis ...
        wvl = np.linspace(1212.,1216.,200)
        # ... and evaluate model
        m = v.evaluate(wvl)
        
        # Add (Gaussian) instrumental broadening with resolution 5000
        v["R"] = 5000
        mr = v.evaluate(wvl)
        
        # Plot result
#         plt.plot(wvl, m, 'b.-', label="R = inf")
#         plt.plot(wvl, mr, 'r.-', label="R = 5000")
#         plt.legend()
#         plt.show()

    def testsanity_LyATransmission(self):
        """
        Checking sanity of LyATransmission example
        """
        from PyAstronomy import modelSuite as ms
        import numpy as np
        import matplotlib.pylab as plt

        la = ms.LyaTransmission()
        # Set some parameters
        la["N"] = 5e17
        la["b"] = 12.2
        la["Dfrac"] = 1.9e-5

        # Set up wavelength axis ...
        wvl = np.linspace(1214., 1217., 1000)
        # ... and evaluate model
        m = la.evaluate(wvl)

        # Plot the result
        plt.plot(wvl, m, 'b.-')
#    plt.show()

    def testsanity_RotBroadProfileExample(self):
        """
        Example of rotational broadening.
        """
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import modelSuite as ms

        # Get an instance of the model ...
        x = ms.RotBroadProfile()
        # ... and define some starting value
        x["xmax"] = 60.0
        x["A"] = 1.0
        x["eps"] = 0.8
        x["off"] = 0.0

        # Define a radial velocity axis
        vv = np.linspace(-90., 90., 200)

        # Construct some "data" and ...
        data = x.evaluate(vv)
        # ... add noise
        data += np.random.normal(0.0, 1e-3, data.size)

        # Fit the model using A, xmax, and eps as free
        # parameters ...
        x.thaw(["A", "xmax", "eps"])
        x.fit(vv, data)
        # ... and show the resulting parameter values.
        x.parameterSummary()

        # Plot the data and the model
        plt.plot(vv, data, 'bp')
        plt.plot(vv, x.model, 'r--')
#    plt.show()

    def testsanity_RotBroadProfile(self):
        """
        Checking RotBroadProfile
        """
        import numpy as np
        from PyAstronomy import modelSuite as ms
        import scipy.integrate as sci

        # Get an instance of the model ...
        x = ms.RotBroadProfile()
        vv = np.linspace(-90., 90., 200)

        for i in range(10):
            # ... and define some starting value
            x["xmax"] = np.random.random() * 50.0 + 30.0
            x["A"] = np.random.random() * 10.0 + 1.0
            x["eps"] = np.random.random()

            d = x.evaluate(vv)
            a = sci.trapezoid(d, vv)
            self.assertAlmostEqual(x["A"], a, delta=1.0 / 200., msg="Incorrect profile normalization (" +
                                   "%g vs %g)" % (x["A"], a))

        x["eps"] = 0.0
        x["xmax"] = 50.0
        x["A"] = 1.0
        vv = np.linspace(-x["xmax"], x["xmax"], 100)
        d = x.evaluate(vv)

        y = d - 2.0 / (np.pi * x["xmax"]) * np.sqrt(1.0 - (vv / x["xmax"])**2)
        self.assertFalse(np.any(np.abs(y) > 1e-6),
                         msg="Incorrect profile for eps=0.0")


    def testsanity_KeplerRVModel_example(self):
        """
        Checking sanity of KeplerRVModel example
        """
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy.modelSuite import KeplerRVModel
        from PyAstronomy import funcFit as fuf
        
        # Generate artificial data ...
        jd = np.arange(100)
        rv = 1.5 * np.sin(jd / 37.6 * 2.*np.pi)
        # ... with some error
        rverr = 0.5
        rv += np.random.normal(0, rverr, len(jd))
        rverr = np.ones(len(rv)) * rverr
        
        # Get RV model with one planet (mp) and a potential constant offset
        # in RV (deg = 0)
        krvm = KeplerRVModel(mp=1, deg=0)
        # To obtain some useful estimate of the minimum mass of the companion,
        # we must specify the mass of the star (in terms of solar masses)
        krvm["mstar"] = 0.5
        
        # Let us have a look at the available parameters.
        # Note that not all are meant for fitting in this model (MA and a)!
        # There is also not much use in fitting 'mstar'. It may, however, be
        # used in combination with a prior to take into account its uncertainty in
        # the estimates.
        krvm.parameterSummary(sorting="ps")
        
        # We specify some guess parameters. 
        krvm["per1"] = 37.0
        krvm["K1"] = 1.0
        krvm["e1"] = 0.0
        krvm["tau1"] = 17.0
        krvm["w1"] = 180.
        
        # Let us fit all of these but period ...
        krvm.thaw(["K1", "tau1", "w1", "e1", "c0"])
        # ... and now also the period
        krvm.thaw(["per1"])
        krvm.fit(jd, rv, yerr=rverr)
        # and then get the best-fit model
        kmo = krvm.evaluate(jd)
        
        # What about chi-square and RMS?
        chi = np.sum( (rv - krvm.model)**2 / rverr**2 )
        # Reduced chi-square
        rchi = chi / (len(rv) - len(krvm.freeParameters()))
        print("chi-square and reduced chi-square: %6.3f, %6.3f" % (chi, rchi))
        rms = np.std(rv - krvm.model)
        print("RMS: ", rms)
        
        plt.title("RV data (blue) and model (red)")
        plt.errorbar(jd, rv, yerr=rverr, fmt='b+')
        plt.plot(jd, krvm.model, 'r-')
        #=======================================================================
        # plt.show()
        #=======================================================================
        
        
        
        # Now let us do some posterior-based error analysis using MCMC
        
        # Say, we want 20 burn-in iterations and, thereafter,
        # 50 further iterations (per walker).
        sampleArgs = {"iters":50, "burn":100}
        
        # Specify a bounded uniform prior on the eccentricity. Note that restrictions are not
        # automatically converted into priors (they may not ne uniform). Potentially further prior,
        # e.g., on per1 may be required to prevent wandering into 'forbidden territory'.
        priors = {"e1":fuf.FuFPrior("limuniform", upper=1, lower=0)}
        
        # Start the sampling (ps could be used to continue the sampling)
        ps = krvm.fitEMCEE(jd, rv, yerr=rverr, sampleArgs=sampleArgs, scales={"e":0.05}, dbfile="chain1.emcee", \
            priors=priors)
        
        
        # Have a look at the posterior
        ta = fuf.TraceAnalysis("chain1.emcee")
        
        # What about the deviance (-2 log(Likelihood))
        ta.plotTraceHist("deviance")
        #=======================================================================
        # ta.show()
        #=======================================================================
        
        # Expectation value and highest probability density interval for eccentricity
        ta.plotTraceHist("e1")
        print("Expectation value for eccentricity: ", ta.mean("e1"))
        print("90% HPD for eccentricity: ", ta.hpd("e1", cred=0.9))
        #=======================================================================
        # ta.show()
        #=======================================================================
        


class TestVoigtAstroPSanity(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass
    
    def testsanity_normalization(self):
        """
        Normalization of AstroVoigtP
        """
        import numpy as np
        from PyAstronomy import modelSuite as ms
        from PyAstronomy import pyaC
        
        v = ms.VoigtAstroP()
        
        # Hypothetical wvls [A]
        w0s = [1000., 5000, 10000.]
        
        # (pi e**2)/(m_e c)
        const = (4.803e-10)**2*np.pi / (9.11e-28*29979245800.0)
        
        fs = [1, 100]
        
        for f in fs:
            for w0 in w0s:
        
                v["w0"] = w0
                v["b"] = 100.
                v["gamma"] = 1e-10
                v["f"] = f
        
                dw = 20.0
                w = np.linspace(w0-dw, w0+dw, 1000)
        
                m = v.evaluate(w)
        
                i = pyaC.ibtrapz(w/1e8, m*29979245800.0/(w/1e8)**2 , (w0-dw)/1e8, (w0+dw)/1e8)
                
                self.assertAlmostEqual(i/const, f, delta=1e-2, msg="Normalization of AstroVoigtP is broken: f, w0, i: % g, % g, % g" % (f, w0, i))

    def testsanity_instrumentalResolution(self):
        """
        Checking integrity of instrumental resolution in VoigtAstroP
        """
        import numpy as np
        from PyAstronomy import modelSuite as ms
        from PyAstronomy import pyasl
        
        v = ms.VoigtAstroP()
        
        w0 = 10830
        
        for R in [2500, 5000, 80000]:
        
            v["w0"] = w0
            v["b"] = 10.
            v["gamma"] = 1e-8
            v["f"] = 100.0
            v["R"] = 0
        
            dw = 40.0
            w = np.linspace(w0-dw, w0+dw, 4000)
        
            m = v.evaluate(w)
        
            v["R"] = R
            m2 = v.evaluate(w)
        
            
            fb = pyasl.instrBroadGaussFast(w, m, R, edgeHandling=None, fullout=False, maxsig=None)

            d = 1000
            
            self.assertAlmostEqual(np.max(np.abs(fb[d:-d] - m2[d:-d])/m2[d:-d]), 0.0, delta=1e-10,
                                   msg="VoigtAstroP instrumental broadening broken for R = " + str(R))

            



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import datetime
import numpy
from numpy import sin, cos, tan, sqrt, arcsin
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo


def weekday(date):
    """
    Return weekday by name

    Parameters
    ----------
    date : DATETIME object
       Give the date of a day as DATETIME object.

    Returns
    -------
    name: string
        "Monday", "Tuesday", "Wednesday",
        "Thursday", "Friday", "Saturday", "Sunday"
    """
    if not isinstance(date, datetime.date):
        raise(ValueError("ERROR weekday: Given date is not a DATETIME object."))

    dayname = ["Monday", "Tuesday", "Wednesday",
               "Thursday", "Friday", "Saturday", "Sunday"]
    ind = date.weekday()

    return dayname[ind]


def daycnv(xjd, mode="idl"):
    """
    Converts Julian dates into Gregorian calendar dates. 

    Handles both individual floats as xjd and iterables such as
    lists and arrays. In the latter case, the result is returned
    in the form of a list.

    Parameters
    ----------
    xjd : float, list, array
        The Julian date
    mode : string, {idl, dtlist, dt}, optional
        Determines format of output. If 'idl' is given (default),
        a list holding [year, month, day, (fractional) hours] is
        returned; this mimics the behavior of the IDL astrolib function.
        If 'dtlist' is given, a list holding
        [year, month, day, hours, minutes, seconds, microseconds] is
        returned. Finally, if 'dt' is specified, a Python
        datetime object will be returned. If the input is an iterable,
        the mode determines the format of the individual items in the
        result list.

    Returns
    -------
    Calendar date : list or datetime object
        A list holding [year, month, day, (fractional) hours] (default)
        or [year, month, day, hours, minutes, seconds, microseconds].
        Alternatively, a Python datetime object is returned. The format
        depends on the 'mode' specified. If the input is an iterable of
        Julian dates, the output is a list.

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.
    :IDL - Documentation:
    NAME:
          DAYCNV
    PURPOSE:
          Converts Julian dates to Gregorian calendar dates
    CALLING SEQUENCE:
          DAYCNV, XJD, YR, MN, DAY, HR
    INPUTS:
          XJD = Julian date, positive double precision scalar or vector
    OUTPUTS:
          YR = Year (Integer)
          MN = Month (Integer)
          DAY = Day (Integer)
          HR = Hours and fractional hours (Real).   If XJD is a vector,
                  then YR,MN,DAY and HR will be vectors of the same length.
    EXAMPLE:
          IDL> DAYCNV, 2440000.D, yr, mn, day, hr    
          yields yr = 1968, mn =5, day = 23, hr =12.   
    WARNING:
          Be sure that the Julian date is specified as double precision to
          maintain accuracy at the fractional hour level.
    METHOD:
          Uses the algorithm of Fliegel and Van Flandern (1968) as reported in
          the "Explanatory Supplement to the Astronomical Almanac" (1992), p. 604
          Works for all Gregorian calendar dates with XJD > 0, i.e., dates after
          -4713 November 23.
    REVISION HISTORY:
          Converted to IDL from Yeoman's Comet Ephemeris Generator, 
          B. Pfarr, STX, 6/16/88
          Converted to IDL V5.0   W. Landsman   September 1997
    """

    if not mode in ('idl', 'dtlist', 'dt'):
        raise(PE.PyAValError("Unknown mode: " + str(mode),
                             where="daycnv",
                             solution="Use any of 'idl', 'dtlist', or 'dt'."))

    # Adjustment needed because Julian day starts at noon, calendar day at midnight

    iterable = hasattr(xjd, "__iter__")

    # Use iterable throughout calculations
    if not iterable:
        xjd = [xjd]

    jd = numpy.array(xjd).astype(int)  # Truncate to integral day
    frac = numpy.array(xjd).astype(float) - jd + \
        0.5  # Fractional part of calendar day
    gi = numpy.where(frac >= 1.0)
    frac[gi] -= 1.0
    jd[gi] += 1

    hr = frac * 24.0
    l = jd + 68569
    n = 4 * l // 146097
    l = l - (146097 * n + 3) // 4
    yr = 4000 * (l + 1) // 1461001
    l = l - 1461 * yr // 4 + 31  # 1461 = 365.25 * 4
    mn = 80 * l // 2447
    day = l - 2447 * mn // 80
    l = mn // 11
    mn = mn + 2 - 12 * l
    yr = 100 * (n - 49) + yr + l
    if mode in ('dt', 'dtlist'):
        # [year, month, day, hours, minutes, seconds, microseconds] requested
        hour = numpy.floor(hr).astype(int)
        minute = numpy.floor((hr - numpy.floor(hr)) * 60).astype(int)
        sec = numpy.floor((hr - hour - minute / 60.) * 3600.).astype(int)
        msec = (3600 * 1e6 * (hr - hour - minute / 60. - sec / 3600.)).astype(int)
        if mode == 'dtlist':
            if not iterable:
                return [yr[0], mn[0], day[0], hour[0], minute[0], sec[0], msec[0]]
            return [[yr[i], mn[i], day[i], hour[i], minute[i], sec[i], msec[i]] for i in smo.range(len(yr))]
        # Return datetime object
        dts = [datetime.datetime(*(yr[i], mn[i], day[i], hour[i],
                                   minute[i], sec[i], msec[i])) for i in smo.range(len(yr))]
        if not iterable:
            return dts[0]
        return dts
    if not iterable:
        return [yr[0], mn[0], day[0], hr[0]]
    return [[yr[i], mn[i], day[i], hr[i]] for i in smo.range(len(yr))]


def bprecess(ra, dec, mu_radec=None,
             parallax=0.0,  rad_vel=0.0, epoch=2000.0):
    """
    Precess positions from J2000.0 (FK5) to B1950.0 (FK4).

    Parameters
    ----------
    ra : float
        Right ascension [deg]
    dec : float
        Declination [deg]
    mu_radec : list
        List of two float entries, optional
        Proper motion [arcsec per tropical CENTURY]
    parallax : float
        The parallax of the target
    rad_vel : float
        Radial velocity [km/s]

    Returns
    -------
    Precessed position : list
        [ra_1950, dec_1950, MU_RADEC, PARALLAX, RAD_VEL] referring to 1950

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
          BPRECESS
    PURPOSE:
          Precess positions from J2000.0 (FK5) to B1950.0 (FK4)
    EXPLANATION:
          Calculates the mean place of a star at B1950.0 on the FK4 system from
          the mean place at J2000.0 on the FK5 system.    

    CALLING SEQUENCE:
          bprecess, ra, dec, ra_1950, dec_1950, [ MU_RADEC = , PARALLAX = 
                                          RAD_VEL =, EPOCH =   ]

    INPUTS:
          RA,DEC - Input J2000 right ascension and declination in *degrees*.
                  Scalar or N element vector

    OUTPUTS:
          RA_1950, DEC_1950 - The corresponding B1950 right ascension and 
                  declination in *degrees*.    Same number of elements as
                  RA,DEC but always double precision.

    OPTIONAL INPUT-OUTPUT KEYWORDS
          MU_RADEC - 2xN element double precision vector containing the proper 
                    motion in seconds of arc per tropical *century* in right 
                    ascension and declination.
          PARALLAX - N_element vector giving stellar parallax (seconds of arc)
          RAD_VEL  - N_element vector giving radial velocity in km/s

          The values of MU_RADEC, PARALLAX, and RADVEL will all be modified
          upon output to contain the values of these quantities in the
          B1950 system.  The parallax and radial velocity will have a very 
          minor influence on the B1950 position.   

          EPOCH - scalar giving epoch of original observations, default 2000.0d
              This keyword value is only used if the MU_RADEC keyword is not set.
    NOTES:
          The algorithm is taken from the Explanatory Supplement to the 
          Astronomical Almanac 1992, page 186.
          Also see Aoki et al (1983), A&A, 128,263

          BPRECESS distinguishes between the following two cases:
          (1) The proper motion is known and non-zero
          (2) the proper motion is unknown or known to be exactly zero (i.e.
                  extragalactic radio sources).   In this case, the reverse of 
                  the algorithm in Appendix 2 of Aoki et al. (1983) is used to 
                  ensure that the output proper motion is  exactly zero. Better 
                  precision can be achieved in this case by inputting the EPOCH 
                  of the original observations.

          The error in using the IDL procedure PRECESS for converting between
          B1950 and J1950 can be up to 12", mainly in right ascension.   If
          better accuracy than this is needed then BPRECESS should be used.

          An unsystematic comparison of BPRECESS with the IPAC precession 
          routine (http://nedwww.ipac.caltech.edu/forms/calculator.html) always 
          gives differences less than 0.15".
    EXAMPLE:
          The SAO2000 catalogue gives the J2000 position and proper motion for
          the star HD 119288.   Find the B1950 position. 

          RA(2000) = 13h 42m 12.740s      Dec(2000) = 8d 23' 17.69''  
          Mu(RA) = -.0257 s/yr      Mu(Dec) = -.090 ''/yr

          IDL> mu_radec = 100D* [ -15D*.0257, -0.090 ]
          IDL> ra = ten(13, 42, 12.740)*15.D 
          IDL> dec = ten(8, 23, 17.69)
          IDL> bprecess, ra, dec, ra1950, dec1950, mu_radec = mu_radec
          IDL> print, adstring(ra1950, dec1950,2)
                  ===> 13h 39m 44.526s    +08d 38' 28.63"

    REVISION HISTORY:
          Written,    W. Landsman                October, 1992
          Vectorized, W. Landsman                February, 1994
          Treat case where proper motion not known or exactly zero  November 1994
          Handling of arrays larger than 32767   Lars L. Christensen, march, 1995
          Converted to IDL V5.0   W. Landsman   September 1997
          Fixed bug where A term not initialized for vector input 
              W. Landsman        February 2000
    """

    radeg = 180.0 / numpy.pi
    sec_to_radian = 1.0 / radeg / 3600.0

    M = numpy.array([[+0.9999256795, -0.0111814828, -0.0048590040,
                      -0.000551,  -0.238560,     +0.435730],
                     [+0.0111814828, +0.9999374849, -0.0000271557,
                      +0.238509,     -0.002667,      -0.008541],
                     [+0.0048590039, -0.0000271771, +0.9999881946,
                      -0.435614,      +0.012254,      +0.002117],
                     [-0.00000242389840, +0.00000002710544, +0.00000001177742,
                      +0.99990432,    -0.01118145,    -0.00485852],
                     [-0.00000002710544, -0.00000242392702, +0.00000000006585,
                      +0.01118145,     +0.99991613,    -0.00002716],
                     [-0.00000001177742, +0.00000000006585, -0.00000242404995,
                      +0.00485852,   -0.00002717,    +0.99996684]])

    # in arc seconds per century
    A_dot = 1e-3 * numpy.array([1.244, -1.579, -0.660])

    ra_rad = ra / radeg
    dec_rad = dec / radeg
    cosra = cos(ra_rad)
    sinra = sin(ra_rad)
    cosdec = cos(dec_rad)
    sindec = sin(dec_rad)

    dec_1950 = 0.0
    ra_1950 = 0.0

    # Following statement moved inside loop in Feb 2000.
    A = 1e-6 * numpy.array([-1.62557, -0.31919, -0.13843])        # in radians

    r0 = numpy.array([cosra * cosdec, sinra * cosdec, sindec])

    if mu_radec is not None:
        mu_a = mu_radec[0]
        mu_d = mu_radec[1]
        # Velocity vector
        r0_dot = numpy.array([-mu_a * sinra * cosdec - mu_d * cosra * sindec,
                              mu_a * cosra * cosdec - mu_d * sinra * sindec,
                              mu_d * cosdec]) + 21.095 * rad_vel * parallax * r0
    else:
        r0_dot = numpy.zeros(3)

    R_0 = numpy.concatenate([r0, r0_dot])
    R_1 = numpy.dot(numpy.transpose(M), R_0)

    # Include the effects of the E-terms of aberration to form r and r_dot.

    r1 = R_1[0:3]
    r1_dot = R_1[3:6]

    if mu_radec is None:
        r1 = r1 + sec_to_radian * r1_dot * (epoch - 1950.0) / 100.
        A = A + sec_to_radian * A_dot * (epoch - 1950.0) / 100.

    x1 = R_1[0]
    y1 = R_1[1]
    z1 = R_1[2]
    rmag = sqrt(x1**2 + y1**2 + z1**2)

    s1 = r1 / rmag
    s1_dot = r1_dot / rmag

    s = s1
    for j in smo.range(2):
        r = s1 + A - ((s * A).sum()) * s
        s = r / rmag

    x = r[0]
    y = r[1]
    z = r[2]
    r2 = x**2 + y**2 + z**2
    rmag = sqrt(r2)

    if mu_radec is not None:
        r_dot = s1_dot + A_dot - ((s * A_dot).sum()) * s
        x_dot = r_dot[0]
        y_dot = r_dot[1]
        z_dot = r_dot[2]
        mu_radec[0] = (x * y_dot - y * x_dot) / (x**2 + y**2)
        mu_radec[1] = (z_dot * (x**2 + y**2) - z *
                       (x * x_dot + y * y_dot)) / (r2 * sqrt(x**2 + y**2))

    dec_1950 = arcsin(z / rmag)
    ra_1950 = numpy.arctan2(y, x)

    if parallax > 0.0:
        rad_vel = (x * x_dot + y * y_dot + z * z_dot) / \
            (21.095 * parallax * rmag)
        parallax = parallax / rmag

    if ra_1950 < 0.0:
        ra_1950 += 2.0 * numpy.pi

    ra_1950 = ra_1950 * radeg
    dec_1950 = dec_1950 * radeg

    return [ra_1950, dec_1950, mu_radec, parallax, rad_vel]


def premat(equinox1, equinox2, FK4=False):
    """
    Return the precession matrix needed to go from EQUINOX1 to EQUINOX2.

    Parameters
    ----------
    equinox1, equinox2 : float
        EQUINOX1 and EQUINOX2
    FK4 : boolean
        Set this to True to obtain output in FK4 system

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
          PREMAT
    PURPOSE:
          Return the precession matrix needed to go from EQUINOX1 to EQUINOX2.  
    EXPLANTION:
          This matrix is used by the procedures PRECESS and BARYVEL to precess 
          astronomical coordinates

    CALLING SEQUENCE:
          matrix = PREMAT( equinox1, equinox2, [ /FK4 ] )

    INPUTS:
          EQUINOX1 - Original equinox of coordinates, numeric scalar.  
          EQUINOX2 - Equinox of precessed coordinates.

    OUTPUT:
        matrix - double precision 3 x 3 precession matrix, used to precess
                  equatorial rectangular coordinates

    OPTIONAL INPUT KEYWORDS:
          /FK4   - If this keyword is set, the FK4 (B1950.0) system precession
                  angles are used to compute the precession matrix.   The 
                  default is to use FK5 (J2000.0) precession angles

    EXAMPLES:
          Return the precession matrix from 1950.0 to 1975.0 in the FK4 system

          IDL> matrix = PREMAT( 1950.0, 1975.0, /FK4)

    PROCEDURE:
          FK4 constants from "Computational Spherical Astronomy" by Taff (1983), 
          p. 24. (FK4). FK5 constants from "Astronomical Almanac Explanatory
          Supplement 1992, page 104 Table 3.211.1.

    REVISION HISTORY
          Written, Wayne Landsman, HSTX Corporation, June 1994
          Converted to IDL V5.0   W. Landsman   September 1997
    """

    deg_to_rad = numpy.pi / 180.0
    sec_to_rad = deg_to_rad / 3600.0

    t = 0.001 * (equinox2 - equinox1)

    if not FK4:
        st = 0.001 * (equinox1 - 2000.0)
        # Compute 3 rotation angles
        A = sec_to_rad * t * (23062.181 + st * (139.656 + 0.0139 * st)
                              + t * (30.188 - 0.344 * st + 17.998 * t))

        B = sec_to_rad * t * t * (79.280 + 0.410 * st + 0.205 * t) + A

        C = sec_to_rad * t * (20043.109 - st * (85.33 + 0.217 * st)
                              + t * (-42.665 - 0.217 * st - 41.833 * t))

    else:
        st = 0.001 * (equinox1 - 1900.0)
        # Compute 3 rotation angles

        A = sec_to_rad * t * (23042.53 + st * (139.75 + 0.06 * st)
                              + t * (30.23 - 0.27 * st + 18.0 * t))

        B = sec_to_rad * t * t * (79.27 + 0.66 * st + 0.32 * t) + A

        C = sec_to_rad * t * (20046.85 - st * (85.33 + 0.37 * st)
                              + t * (-42.67 - 0.37 * st - 41.8 * t))

    sina = sin(A)
    sinb = sin(B)
    sinc = sin(C)
    cosa = cos(A)
    cosb = cos(B)
    cosc = cos(C)

    r = numpy.zeros((3, 3))
    r[::, 0] = numpy.array([cosa * cosb * cosc - sina *
                            sinb, sina * cosb + cosa * sinb * cosc,  cosa * sinc])
    r[::, 1] = numpy.array([-cosa * sinb - sina * cosb *
                            cosc, cosa * cosb - sina * sinb * cosc, -sina * sinc])
    r[::, 2] = numpy.array([-cosb * sinc, -sinb * sinc, cosc])
    return r


def precess(ra, dec, equinox1, equinox2, FK4=False, radian=False):
    """
    Precess coordinates from EQUINOX1 to EQUINOX2.

    Parameters
    ----------
    ra, dec, equinox1, equinox2 : float
        Position and equinox
    FK4 : boolean
        Set to True to obtain output in FK4 system.
    radian : boolean
        If True, `ra` and `dec` must be given in radian (degrees otherwise). 

    Returns
    -------
    Position : list of ra and dec
        A list with [ra, dec] precessed from equinox 1 to equinox 2.

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
        PRECESS
    PURPOSE:
        Precess coordinates from EQUINOX1 to EQUINOX2.  
    EXPLANATION:
        For interactive display, one can use the procedure ASTRO which calls 
        PRECESS or use the /PRINT keyword.   The default (RA,DEC) system is 
        FK5 based on epoch J2000.0 but FK4 based on B1950.0 is available via 
        the /FK4 keyword.

        Use BPRECESS and JPRECESS to convert between FK4 and FK5 systems
    CALLING SEQUENCE:
        PRECESS, ra, dec, [ equinox1, equinox2, /PRINT, /FK4, /RADIAN ]

    INPUT - OUTPUT:
        RA - Input right ascension (scalar or vector) in DEGREES, unless the 
                /RADIAN keyword is set
        DEC - Input declination in DEGREES (scalar or vector), unless the 
                /RADIAN keyword is set

        The input RA and DEC are modified by PRECESS to give the 
        values after precession.

    OPTIONAL INPUTS:
        EQUINOX1 - Original equinox of coordinates, numeric scalar.  If 
                  omitted, then PRECESS will query for EQUINOX1 and EQUINOX2.
        EQUINOX2 - Equinox of precessed coordinates.

    OPTIONAL INPUT KEYWORDS:
        /PRINT - If this keyword is set and non-zero, then the precessed
                  coordinates are displayed at the terminal.    Cannot be used
                  with the /RADIAN keyword
        /FK4   - If this keyword is set and non-zero, the FK4 (B1950.0) system
                  will be used otherwise FK5 (J2000.0) will be used instead.
        /RADIAN - If this keyword is set and non-zero, then the input and 
                  output RA and DEC vectors are in radians rather than degrees

    RESTRICTIONS:
          Accuracy of precession decreases for declination values near 90 
          degrees.  PRECESS should not be used more than 2.5 centuries from
          2000 on the FK5 system (1950.0 on the FK4 system).

    EXAMPLES:
          (1) The Pole Star has J2000.0 coordinates (2h, 31m, 46.3s, 
                  89d 15' 50.6"); compute its coordinates at J1985.0

          IDL> precess, ten(2,31,46.3)*15, ten(89,15,50.6), 2000, 1985, /PRINT

                  ====> 2h 16m 22.73s, 89d 11' 47.3"

          (2) Precess the B1950 coordinates of Eps Ind (RA = 21h 59m,33.053s,
          DEC = (-56d, 59', 33.053") to equinox B1975.

          IDL> ra = ten(21, 59, 33.053)*15
          IDL> dec = ten(-56, 59, 33.053)
          IDL> precess, ra, dec ,1950, 1975, /fk4

    PROCEDURE:
          Algorithm from Computational Spherical Astronomy by Taff (1983), 
          p. 24. (FK4). FK5 constants from "Astronomical Almanac Explanatory
          Supplement 1992, page 104 Table 3.211.1.

    PROCEDURE CALLED:
          Function PREMAT - computes precession matrix 

    REVISION HISTORY
          Written, Wayne Landsman, STI Corporation  August 1986
          Correct negative output RA values   February 1989
          Added /PRINT keyword      W. Landsman   November, 1991
          Provided FK5 (J2000.0)  I. Freedman   January 1994
          Precession Matrix computation now in PREMAT   W. Landsman June 1994
          Added /RADIAN keyword                         W. Landsman June 1997
          Converted to IDL V5.0   W. Landsman   September 1997
          Correct negative output RA values when /RADIAN used    March 1999 
          Work for arrays, not just vectors  W. Landsman    September 2003 
    """
    deg_to_rad = numpy.pi / 180.0

    if not radian:
        # ra, dec are given in degrees
        ra_rad = ra * deg_to_rad     # Convert to double precision if not already
        dec_rad = dec * deg_to_rad
    else:
        ra_rad = ra
        dec_rad = dec

    a = cos(dec_rad)

    x = [a * cos(ra_rad), a * sin(ra_rad), sin(dec_rad)]  # input direction

    sec_to_rad = deg_to_rad / 3600.0

    # Use PREMAT function to get precession matrix from Equinox1 to Equinox2

    r = premat(equinox1, equinox2, FK4=FK4)

    x2 = numpy.dot(r, x)      # rotate to get output direction cosines

    ra_rad = numpy.arctan2(x2[1], x2[0])
    dec_rad = arcsin(x2[2])

    if not radian:
        ra = ra_rad / deg_to_rad
        # RA between 0 and 360 degrees
        ra = ra + int(ra < 0.0) * 360.0
        dec = dec_rad / deg_to_rad
    else:
        ra = ra_rad
        dec = dec_rad
        ra = ra + int(ra < 0.0) * 2.0 * numpy.pi

    return [ra, dec]


def precess_xyz(x, y, z, equinox1, equinox2):
    """
    Precess equatorial geocentric rectangular coordinates. 

    Parameters
    ----------
    x, y, z, equinox1, equinox2 : float

    Returns
    -------
    Precessed coordinates : list
        A list containing the updated `x`, `y`, and `z` values. 

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
          PRECESS_XYZ

    PURPOSE:
          Precess equatorial geocentric rectangular coordinates. 

    CALLING SEQUENCE:
          precess_xyz, x, y, z, equinox1, equinox2

    INPUT/OUTPUT:
          x,y,z: scalars or vectors giving heliocentric rectangular coordinates
                THESE ARE CHANGED UPON RETURNING.
    INPUT:
          EQUINOX1: equinox of input coordinates, numeric scalar
          EQUINOX2: equinox of output coordinates, numeric scalar

    OUTPUT:
          x,y,z are changed upon return

    NOTES:
      The equatorial geocentric rectangular coords are converted
        to RA and Dec, precessed in the normal way, then changed
        back to x, y and z using unit vectors.

    EXAMPLE:
          Precess 1950 equinox coords x, y and z to 2000.
          IDL> precess_xyz,x,y,z, 1950, 2000

    HISTORY:
          Written by P. Plait/ACC March 24 1999 
            (unit vectors provided by D. Lindler)
          Use /Radian call to PRECESS     W. Landsman     November 2000
          Use two parameter call to ATAN   W. Landsman    June 2001
    """

    # take input coords and convert to ra and dec (in radians)
    ra = numpy.arctan2(y, x)
    delp = sqrt(x * x + y * y + z * z)  # magnitude of distance to Sun
    dec = arcsin(z / delp)

    # precess the ra and dec
    ra, dec = precess(ra, dec, equinox1, equinox2, radian=True)

    # convert back to x, y, z
    xunit = cos(ra) * cos(dec)
    yunit = sin(ra) * cos(dec)
    zunit = sin(dec)

    x = xunit * delp
    y = yunit * delp
    z = zunit * delp

    return [x, y, z]


def xyz(date, velocity=False, equinox=1950.0):
    """
    Calculate geocentric X,Y, and Z  and velocity coordinates of the Sun.

    Parameters
    ----------
    date : float
        Julian date
    equinox : float
        Equinox of output. If None, Equinox will be 1950.
    velocity : boolean
        If False, the velocity of the Sun will not be calculated

    Returns
    -------
    Sun position and velocity : list
        A list of the from [X, Y, Z, XVEL, YVEL, ZVEL]. Last three values are None
        if `velocity` flag is set to False.

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
          XYZ
    PURPOSE:
          Calculate geocentric X,Y, and Z  and velocity coordinates of the Sun
    EXPLANATION:
          Calculates geocentric X,Y, and Z vectors and velocity coordinates 
          (dx, dy and dz) of the Sun.   (The positive X axis is directed towards 
          the equinox, the y-axis, towards the point on the equator at right 
          ascension 6h, and the z axis toward the north pole of the equator).
          Typical position accuracy is <1e-4 AU (15000 km).

    CALLING SEQUENCE:
          XYZ, date, x, y, z, [ xvel, yvel, zvel, EQUINOX = ]

    INPUT:
          date: reduced julian date (=JD - 2400000), scalar or vector

    OUTPUT:
          x,y,z: scalars or vectors giving heliocentric rectangular coordinates
                    (in A.U) for each date supplied.    Note that sqrt(x^2 + y^2
                    + z^2) gives the Earth-Sun distance for the given date.
          xvel, yvel, zvel: velocity vectors corresponding to X, Y and Z.

    OPTIONAL KEYWORD INPUT:
          EQUINOX: equinox of output. Default is 1950.

    EXAMPLE:
          What were the rectangular coordinates and velocities of the Sun on 
          Jan 22, 1999 0h UT (= JD 2451200.5) in J2000 coords? NOTE:
          Astronomical Almanac (AA) is in TDT, so add 64 seconds to 
          UT to convert.

          IDL> xyz,51200.5+64.d/86400.d,x,y,z,xv,yv,zv,equinox = 2000

          Compare to Astronomical Almanac (1999 page C20)
                      X  (AU)        Y  (AU)     Z (AU)
          XYZ:      0.51456871   -0.76963263  -0.33376880
          AA:       0.51453130   -0.7697110   -0.3337152
          abs(err): 0.00003739    0.00007839   0.00005360
          abs(err)
              (km):   5609          11759         8040 

          NOTE: Velocities in AA are for Earth/Moon barycenter
                (a very minor offset) see AA 1999 page E3
                    X VEL (AU/DAY) YVEL (AU/DAY)   Z VEL (AU/DAY)
          XYZ:      -0.014947268   -0.0083148382    -0.0036068577
          AA:       -0.01494574    -0.00831185      -0.00360365
          abs(err):  0.000001583    0.0000029886     0.0000032077
          abs(err)
          (km/sec): 0.00265        0.00519          0.00557

    PROCEDURE CALLS:
          PRECESS_XYZ
    REVISION HISTORY
          Original algorithm from Almanac for Computers, Doggett et al. USNO 1978
          Adapted from the book Astronomical Photometry by A. Henden
          Written  W. Landsman   STX       June 1989
          Correct error in X coefficient   W. Landsman HSTX  January 1995
          Added velocities, more terms to positions and EQUINOX keyword,
            some minor adjustments to calculations 
            P. Plait/ACC March 24, 1999
    """

    picon = numpy.pi / 180.0
    t = (date - 15020.0) / 36525.0         # Relative Julian century from 1900

    # NOTE: longitude arguments below are given in *equinox* of date.
    #   Precess these to equinox 1950 to give everything an even footing.
    #   Compute argument of precession from equinox of date back to 1950
    pp = (1.396041 + 0.000308 * (t + 0.5)) * (t - 0.499998)

    # Compute mean solar longitude, precessed back to 1950
    el = 279.696678 + 36000.76892 * t + 0.000303 * t * t - pp

    # Compute Mean longitude of the Moon
    c = 270.434164 + 480960. * t + 307.883142 * t - 0.001133 * t * t - pp

    # Compute longitude of Moon's ascending node
    n = 259.183275 - 1800. * t - 134.142008 * t + 0.002078 * t * t - pp

    # Compute mean solar anomaly
    g = 358.475833 + 35999.04975 * t - 0.00015 * t * t

    # Compute the mean jupiter anomaly
    j = 225.444651 + 2880.0 * t + 154.906654 * t * t

    # Compute mean anomaly of Venus
    v = 212.603219 + 58320. * t + 197.803875 * t + 0.001286 * t * t

    # Compute mean anomaly of Mars
    m = 319.529425 + 19080. * t + 59.8585 * t + 0.000181 * t * t

    # Convert degrees to radians for trig functions
    el = el * picon
    g = g * picon
    j = j * picon
    c = c * picon
    v = v * picon
    n = n * picon
    m = m * picon

    # Calculate X,Y,Z using trigonometric series
    X = 0.999860 * cos(el)                          \
        - 0.025127 * cos(g - el)                      \
        + 0.008374 * cos(g + el)                      \
        + 0.000105 * cos(g + g + el)                  \
        + 0.000063 * t * cos(g - el)                    \
        + 0.000035 * cos(g + g - el)                  \
        - 0.000026 * sin(g - el - j)                  \
        - 0.000021 * t * cos(g + el)                    \
        + 0.000018 * sin(2. * g + el - 2. * v)            \
        + 0.000017 * cos(c)                           \
        - 0.000014 * cos(c - 2. * el)                   \
        + 0.000012 * cos(4. * g + el - 8. * m + 3. * j)     \
        - 0.000012 * cos(4. * g - el - 8. * m + 3. * j)     \
        - 0.000012 * cos(g + el - v)                  \
        + 0.000011 * cos(2. * g + el - 2. * v)            \
        + 0.000011 * cos(2. * g - el - 2. * j)

    Y = 0.917308 * sin(el)                             \
        + 0.023053 * sin(g - el)                         \
        + 0.007683 * sin(g + el)                         \
        + 0.000097 * sin(g + g + el)                     \
        - 0.000057 * t * sin(g - el)                       \
        - 0.000032 * sin(g + g - el)                     \
        - 0.000024 * cos(g - el - j)                     \
        - 0.000019 * t * sin(g + el)                       \
        - 0.000017 * cos(2.00 * g + el - 2.00 * v)           \
        + 0.000016 * sin(c)                              \
        + 0.000013 * sin(c - 2.00 * el)                   \
        + 0.000011 * sin(4.00 * g + el - 8.00 * m + 3.00 * j)  \
        + 0.000011 * sin(4.00 * g - el - 8.00 * m + 3.00 * j)  \
        - 0.000011 * sin(g + el - v)                     \
        + 0.000010 * sin(2.00 * g + el - 2.00 * v)          \
        - 0.000010 * sin(2.00 * g - el - 2.00 * j)

    Z = 0.397825 * sin(el)        \
        + 0.009998 * sin(g - el)      \
        + 0.003332 * sin(g + el)      \
        + 0.000042 * sin(g + g + el)    \
        - 0.000025 * t * sin(g - el)    \
        - 0.000014 * sin(g + g - el)    \
        - 0.000010 * cos(g - el - j)

    # Precess_to new equator?
    if equinox is not None:
        X, Y, Z = precess_xyz(X, Y, Z, 1950, equinox)

    if not velocity:
        return [X, Y, Z, None, None, None]

    XVEL = -0.017200 * sin(el)          \
        - 0.000288 * sin(g + el)       \
        - 0.000005 * sin(2.00 * g + el)  \
        - 0.000004 * sin(c)            \
        + 0.000003 * sin(c - 2.00 * el)  \
        + 0.000001 * t * sin(g + el)      \
        - 0.000001 * sin(2.00 * g - el)

    YVEL = 0.015780 * cos(el)           \
        + 0.000264 * cos(g + el)        \
        + 0.000005 * cos(2.00 * g + el)   \
        + 0.000004 * cos(c)             \
        + 0.000003 * cos(c - 2.00 * el)   \
        - 0.000001 * t * cos(g + el)

    ZVEL = 0.006843 * cos(el)             \
        + 0.000115 * cos(g + el)        \
        + 0.000002 * cos(2.00 * g + el)    \
        + 0.000002 * cos(c)              \
        + 0.000001 * cos(c - 2.00 * el)

    # Precess to new equator?

    if equinox != 1950.0:
        XVEL, YVEL, ZVEL = precess_xyz(XVEL, YVEL, ZVEL, 1950, equinox)

    return [X, Y, Z, XVEL, YVEL, ZVEL]


def helio_jd(date, ra, dec, B1950=False, TIME_DIFF=False):
    """
    Convert geocentric (reduced) Julian date into heliocentric Julian date

    .. note:: The heliocentric julian date (HJD) differs from the barycentric julian date (BJD)
              by +/- 4 s (e.g., Eastman et al. 2010, PASP 122, 935). See also the
              :ref:`timescales`.

    Parameters
    ----------
    date : float
        (Reduced) Julian date (2.4e6 subtracted)
    ra, dec : float
        Right ascension and declination in degrees
    B1950 : boolean
        If True, input coordinates are assumed to be given in equinox
        1950 coordinates.
    TIME_DIFF : boolean
        If True, this function returns the time difference
        (heliocentric JD - geocentric JD ) in seconds

    Returns
    -------
    HJD : float
        The heliocentric Julian date.

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
         HELIO_JD
    PURPOSE:
         Convert geocentric (reduced) Julian date to heliocentric Julian date
    EXPLANATION:
         This procedure correct for the extra light travel time between the Earth 
         and the Sun.

          An online calculator for this quantity is available at 
          http://www.physics.sfasu.edu/astro/javascript/hjd.html
    CALLING SEQUENCE:
          jdhelio = HELIO_JD( date, ra, dec, /B1950, /TIME_DIFF)

    INPUTS
          date - reduced Julian date (= JD - 2400000), scalar or vector, MUST
                  be double precision
          ra,dec - scalars giving right ascension and declination in DEGREES
                  Equinox is J2000 unless the /B1950 keyword is set

    OUTPUTS:
          jdhelio - heliocentric reduced Julian date.  If /TIME_DIFF is set, then
                    HELIO_JD() instead returns the time difference in seconds
                    between the geocentric and heliocentric Julian date.

    OPTIONAL INPUT KEYWORDS 
          /B1950 - if set, then input coordinates are assumed to be in equinox 
                   B1950 coordinates.
          /TIME_DIFF - if set, then HELIO_JD() returns the time difference
                   (heliocentric JD - geocentric JD ) in seconds 

    EXAMPLE:
          What is the heliocentric Julian date of an observation of V402 Cygni
          (J2000: RA = 20 9 7.8, Dec = 37 09 07) taken June 15, 1973 at 11:40 UT?

          IDL> juldate, [1973,6,15,11,40], jd      ;Get geocentric Julian date
          IDL> hjd = helio_jd( jd, ten(20,9,7.8)*15., ten(37,9,7) )  

          ==> hjd = 41848.9881

    Wayne Warren (Raytheon ITSS) has compared the results of HELIO_JD with the
    FORTRAN subroutines in the STARLINK SLALIB library (see 
    http://star-www.rl.ac.uk/).    
                                                     Time Diff (sec)
         Date               RA(2000)   Dec(2000)  STARLINK      IDL

    1999-10-29T00:00:00.0  21 08 25.  -67 22 00.  -59.0        -59.0
    1999-10-29T00:00:00.0  02 56 33.4 +00 26 55.  474.1        474.1
    1940-12-11T06:55:00.0  07 34 41.9 -00 30 42.  366.3        370.2
    1992-02-29T03:15:56.2  12 56 27.4 +42 10 17.  350.8        350.9
    2000-03-01T10:26:31.8  14 28 36.7 -20 42 11.  243.7        243.7
    2100-02-26T09:18:24.2  08 26 51.7 +85 47 28.  104.0        108.8
    PROCEDURES CALLED:
          bprecess, xyz, zparcheck

    REVISION HISTORY:
          Algorithm from the book Astronomical Photometry by Henden, p. 114
          Written,   W. Landsman       STX     June, 1989 
          Make J2000 default equinox, add B1950, /TIME_DIFF keywords, compute
          variation of the obliquity      W. Landsman   November 1999
    """

    # Because XYZ uses default B1950 coordinates, we'll convert everything to B1950

    if date > 2.4e6:
        PE.warn(PE.PyAValError("The given Julian Date ( " + str(date) + ") is exceedingly large far a reduced JD.",
                               solution="Did you forget to subtract 2.4e6?",
                               where="helio_jd"))

    if not B1950:
        bpresult = bprecess(ra, dec)
        ra1 = bpresult[0]
        dec1 = bpresult[1]
    else:
        ra1 = ra
        dec1 = dec

    radeg = 180.0 / numpy.pi
    # I think, this is not needed in Python, even at this stage...
    # zparcheck,'HELIO_JD',date,1,[3,4,5],[0,1],'Reduced Julian Date'

    delta_t = (date - 33282.42345905) / 36525.0
    epsilon_sec = 44.836 - 46.8495 * delta_t - \
        0.00429 * delta_t**2 + 0.00181 * delta_t**3
    epsilon = (23.433333 + epsilon_sec / 3600.0) / radeg
    ra1 = ra1 / radeg
    dec1 = dec1 / radeg

    x, y, z, tmp, tmp, tmp = xyz(date)

    # Find extra distance light must travel in AU, multiply by 1.49598e13 cm/AU,
    # and divide by the speed of light, and multiply by 86400 second/year

    time = -499.00522 * (cos(dec1) * cos(ra1) * x +
                         (tan(epsilon) * sin(dec1) + cos(dec1) * sin(ra1)) * y)

    if TIME_DIFF:
        return time
    else:
        return (date + time / 86400.0)


def jdcnv(dt):
    """
    Converts Gregorian dates to Julian days  

    Parameters
    ----------
    dt : DateTime object
        The date. This is interpreted as UTC and the timezone component is not considered.

    Returns
    -------
    Julian day : float

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

     NAME:
           JDCNV
     PURPOSE:
           Converts Gregorian dates to Julian days   

     EXPLANATION:
           For IDL versions V5.1 or greater, this procedure is superceded by
           JULDAY() function in the standard IDL distribution.   Note, however,
           that prior to V5.1 there wasa bug in JULDAY() that gave answers off
           by 0.5 days. 

     CALLING SEQUENCE:
           JDCNV, YR, MN, DAY, HR, JULIAN

     INPUTS:
           YR = Year, integer scalar or vector
           MN = Month  integer (1-12) scalar or vector
           DAY = Day   integer 1-31) scalar or vector 
           HR  = Hours and fractions of hours of universal time (U.T.), scalar
                  or vector

     OUTPUTS:
           JULIAN = Julian date (double precision) 

     EXAMPLE:
           To find the Julian Date at 1978 January 1, 0h (U.T.)

           IDL> JDCNV, 1978, 1, 1, 0., JULIAN

           will give JULIAN = 2443509.5
     NOTES:
           (1) JDCNV will accept vector arguments 
           (2) JULDATE is an alternate procedure to perform the same function

     REVISON HISTORY:
           Converted to IDL from Don Yeomans Comet Ephemeris Generator,
           B. Pfarr, STX, 6/15/88
           Converted to IDL V5.0   W. Landsman   September 1997
           Added checks on valid month, day ranges W. Landsman July 2008
    """

    # L = leap years, -1 for Jan, Feb, else 0
    L = int((dt.month - 14.0) / 12.0)
    julian = dt.day - 32075 + \
        int(1461 * (dt.year + 4800 + L) / 4.0) + \
        int(367 * (dt.month - 2 - L * 12) / 12.0) - \
        int(int(3 * ((dt.year + 4900 + L) / 100.0)) / 4.0)
    julian += ((dt.hour / 24.0) + (dt.minute / (24.0 * 60.0)) +
               (dt.second / 86400.) + (dt.microsecond / (86400. * 1e6)) - 0.5)

    return julian


def get_juldate():
    """
    Return the current Julian Date

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
        GET_JULDATE
    PURPOSE:
        Return the current Julian Date

    EXPLANATION:
        In V5.4, GET_JULDATE became completely obsolete with the introduction
        of the /UTC keyword to SYSTIME().   So GET_JULDATE,jd is equivalent to
        jd = SYSTIME(/JULIAN,/UTC).

    CALLING SEQUENCE:
          GET_JULDATE,jd

    INPUTS:
          None

    OUTPUTS:
          jd = Current Julian Date, double precision scalar

    EXAMPLE:
          Return the current hour, day, month and year as integers

          IDL> GET_JULDATE, JD                  ;Get current Julian date
          IDL> DAYCNV, JD, YR, MON, DAY, HOURS  ;Convert to hour,day month & year

    METHOD:
          A call is made to SYSTIME(/JULIAN,/UTC).

    REVISION HISTORY:
          Written Wayne Landsman                March, 1991
          Converted to IDL V5.0   W. Landsman   September 1997
          Assume since V5.4 Use /UTC keyword to SYSTIME()  W. Landsman April 2006
    """
    now = datetime.datetime.utcnow()
    return jdcnv(now)


def juldate(date):
    """
    Convert from calendar to Reduced Julian Date

    This function returns the *reduced* Julian date, which
    is obtained by subtracting 2400000 from the Julian date.
    To convert the output into Modified Julian Date (MJD),
    another 0.5 days have to be subtracted.

    Parameters
    ----------
    date : DateTime object
        Calendar date

    Returns
    -------
    RJD : float
        The **reduced** Julian date.

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
      JULDATE
    PURPOSE:
      Convert from calendar to Reduced Julian Date

    EXPLANATION:
      Julian Day Number is a count of days elapsed since Greenwich mean noon 
      on 1 January 4713 B.C.  The Julian Date is the Julian day number
      followed by the fraction of the day elapsed since the preceding noon. 

      This procedure duplicates the functionality of the JULDAY() function in
      in the standard IDL distribution, but also allows interactive input and
      gives output as Reduced Julian date (=JD - 2400000.)  
      (Also note that prior to V5.1 there was a bug in JULDAY() that gave 
      answers offset by 0.5 days.)

    CALLING SEQUENCE:
      JULDATE, /PROMPT           ;Prompt for calendar Date, print Julian Date
                or
      JULDATE, date, jd      

    INPUT:
      DATE -  3 to 6-element vector containing year,month (1-12),day, and 
                optionally hour, minute, and second all specified as numbers
                (Universal Time).   Year should be supplied with all digits.
                Years B.C should be entered as negative numbers (and note that
                Year 0 did not exist).  If Hour, minute or seconds are not 
                supplied, they will default to 0. 

    OUTPUT:
        JD - Reduced Julian date, double precision scalar.  To convert to
                Julian Date, add 2400000.   JULDATE will print the value of
                JD at the terminal if less than 2 parameters are supplied, or 
                if the /PROMPT keyword is set

    OPTIONAL INPUT KEYWORD:
        /PROMPT - If this keyword is set and non-zero, then JULDATE will prompt
                for the calendar date at the terminal.

    RESTRICTIONS:
        The procedure HELIO_JD can be used after JULDATE, if a heliocentric
        Julian date is required.

    EXAMPLE:
        A date of 25-DEC-2006 06:25 UT may be expressed as either

        IDL> juldate, [2006, 12, 25, 6, 25], jd       
        IDL> juldate, [2006, 12, 25.2673611d], jd 

        In either case, one should obtain a Reduced Julian date of 
        JD = 54094.7673611

    PROCEDURE USED:
        GETOPT()
    REVISION HISTORY
        Adapted from IUE RDAF (S. Parsons)                      8-31-87
        Algorithm from Sky and Telescope April 1981   
        Added /PROMPT keyword, W. Landsman    September 1992
        Converted to IDL V5.0   W. Landsman   September 1997
        Make negative years correspond to B.C. (no year 0), work for year 1582
        Disallow 2 digit years.    W. Landsman    March 2000
    """
    jd = jdcnv(date)
    jd -= 2400000.0
    return jd



from __future__ import print_function, division
from PyAstronomy.pyaC import pyaErrors as PE
import os
import six


class AtomicNo:

    def __init__(self):
        """
        """

        path = os.path.dirname(__file__)

        self._data = {}
        self._elNames = {}
        with open(os.path.join(path, "atomicNo.dat")) as df:
            for l in df:
                l = l.rstrip("\n")
                s = l.split()
                self._data[int(s[2])] = s[1]
                self._elNames[int(s[2])] = s[0]
        # Add reverse mapping to _data dict, i.e., element symbol->atomic number
        self._data.update(dict((self._data[k], k) for k in self._data))

    def getElSymbol(self, atn):
        """
        Convert atomic number into elemental symbol.

        Parameters
        ----------
        atn : int
            Atomic number

        Returns
        -------
        Symbol : string
            Elemental symbol
        """
        try:
            atn = int(atn)
        except ValueError:
            raise (PE.PyAValError(
                "`atn` needs to be an integer. You gave: " + str(atn)))

        if not atn in self._data:
            raise (PE.PyAValError(
                "No such atomic number in the data base: " + str(atn)))

        return self._data[atn]

    def getAtomicNo(self, sym):
        """
        Convert elemental symbol into atomic number.

        Parameters
        ----------
        sym : string
            Elemental symbol

        Returns
        -------
        Atomic number : int
            Atomic number
        """
        if not isinstance(sym, six.string_types):
            raise (PE.PyAValError("`sym` must be a string, e.g., 'Fe'."))

        if not sym in self._data:
            raise (PE.PyAValError("No such symbol in the data base: " + str(sym)))

        return self._data[sym]

    def showAll(self):
        """
        Print atomic number, elemental symbol, and element name.
        """
        nos = [k for k in self._data.keys() if type(k) == int]
        for k in sorted(nos):
            if not type(k) == int:
                continue
            print("%3d | %3s | " % (k, self._data[k]) + self._elNames[k])

    def getElementName(self, atn):
        """
        Convert atomic number into elemental name.

        Parameters
        ----------
        atn : int
            Atomic number

        Returns
        -------
        Name : string
            Name of element
        """
        try:
            atn = int(atn)
        except ValueError:
            raise (PE.PyAValError(
                "`atn` needs to be an integer. You gave: " + str(atn)))

        if not atn in self._data:
            raise (PE.PyAValError(
                "No such atomic number in the data base: " + str(atn)))

        return self._elNames[atn]



from PyAstronomy.funcFit import ic

if ic.check["numpy"] and ic.check["scipy"]:
    from .anaMCMCTraces import TraceAnalysis, hpd, quantiles, modeGrenander
    from .bayesFactors import bayesFactors
    __all__ = ["TraceAnalysis", "bayesFactors", "hpd", "quantiles", "modeGrenander"]


# -*- coding: utf-8 -*-
from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np
from PyAstronomy.pyasl import _ic


def planck(T, lam=None, nu=None):
    """
    Evaluate Planck's radiation law.

    Depending on whether wavelength or frequency is specified as input, the
    function evaluates:

    .. math::

        B_{\\nu} = \\frac{2\\pi h \\nu^3}{c^2} \\frac{1}{e^{\\frac{h\\nu}{kT}} - 1}

    or

    .. math::

        B_{\\lambda} = \\frac{2\\pi h c^2}{\\lambda^5} \\frac{1}{e^{\\frac{h c}{\\lambda kT}} - 1} \\; .

    If lambda is given (in meters), the output units are W/(m^2 m). To convert into erg/(cm^2 A s),
    the output has to be multiplied by a factor of 1e-7.

    Parameters
    ----------
    T : float
        Temperature in Kelvin.
    lam : float or array, optional
        Wavelength in meters.
    nu : float or array, optional
        Frequency in Hz.

    Returns
    -------
    Spectral radiance : float or array
        Depending on whether `lam` or `nu` were specified, returns the
        spectral radiance per area and wavelength or frequency. The unit (SI)
        will be W/(m^2 m) if `lam` was given and W/(m^2 Hz) if `nu` was
        specified.
    """
    if _ic.check["quantities"]:
        from PyAstronomy import constants
        c = constants.PyAConstants(unitSystem="SI")
    else:
        raise(PE.PyARequiredImport(
            "You need to install 'quantities' to use this function.\n 'quantities' can be obtained from 'http://pypi.python.org/pypi/quantities'."))
        return None
    if (lam is not None) and (nu is not None):
        raise(PE.PyAParameterConflict(
            "Specify either 'lam' OR 'nu', but not both."))
    if (lam is None) and (nu is None):
        raise(PE.PyAParameterConflict("Specify either 'lam' OR 'nu'."))
    # Parameters have been specified properly
    if lam is not None:
        result = 2.*np.pi*c.h*(c.c**2)/(lam**5)
        result /= (np.exp(c.h*c.c/(lam*c.k*T)) - 1.0)
        return result
    elif nu is not None:
        result = 2.*np.pi*c.h*(nu**3)/(c.c**2)
        result /= (np.exp(c.h*nu/(c.k*T)) - 1.0)
        return result



from __future__ import print_function, division
import numpy as np
import six.moves as smo


def coordinateGrid(*args):
    r"""
    Construct a coordinate array for n-dimensional fitting.

    For n-dimensional fitting, `funcFit` requires a mapping from
    array index (i,j,k,...) to coordinate (x,y,z,...).
    This mapping must be given in the form of an array with
    dimension (n1, n2, ..., nd), where ni is the length of
    the i-th coordinate axis and nd is the number of coordinate
    axes. Then, e.g. in the 2d case, [0,0,0] gives the x-coordinate
    of index [0,0] and [0,0,1] given the associated y-coordinate.

    Parameters
    ----------
    args : arrays
        An arbitrary number of coordinate arrays.

    Returns
    -------
    Coordinate grid : array
        An array of dimension (n1, n2, ..., nd) where
        n1, n2, etc. is the length of the n-th coordinate
        axis and nd is the total number of dimensions.
    """
    # Number of axis
    na = len(args)
    # Length of individual axes
    la = []
    # Number of points in cubes
    npo = 1
    # Shape of output array
    shape = []
    for a in args:
        shape.append(len(a))
        la.append(len(a))
        npo *= la[-1]
    shape.append(na)
    # The resulting output grid
    g = np.zeros(shape)
    # Contains the number of data points in "sub-cubes"
    # Holds: nx, nx*ny, nx*ny*nz, etc.
    npcube = [1]
    for i in smo.range(len(la)):
        npcube.append(npcube[-1]*la[i])
    npcube.pop(0)
    # Loop over all points
    # ai = Array Index
    ai = [0]*na
    for i in smo.range(npo):
        itemp = i
        for j in smo.range(na-1, 0, -1):
            ai[j] = itemp // npcube[j-1]
            itemp -= ai[j]*npcube[j-1]
        ai[0] = itemp
        # Assign the coordinates
        for k in smo.range(na):
            g[tuple(ai)][k] = args[k][ai[k]]
    return g



# -*- coding: utf-8 -*-
from PyAstronomy.funcFit import ic
import os.path
from PyAstronomy.pyaC import pyaErrors as PE

def updateFitsHeader(model, hdr, clobber=False, conf=0.9):
    """
      Update the delivered fits-header with the parameters of the model.

      Parameters:
        - `hdr` - Pyfits header which should be updated with the model parameters.
        - `conf' - Confidence level for MCMC errors.
        - `clobber` - Allows to overwrite parameters which might be already present in the header.
    """
    if not ic.check["pyfits"]:
      raise(PE.PyARequiredImport("pyfits required to use fits file.", where="Params::updateFitsHeader"))
      return
    try:
       # @FIXME The next line can NEVER work
#       x=hdr[p]
       raise(PE.pyaOtherErrors(" Keyword PA_model already present in fits-header, aborting...", where="Params::updateFitsHeader"))
       return
    except:
        pass
    hdr.update('PA_model',model.naming.getRoot(),'PyAstronomy model type')
    for p in model.parameters():
        if clobber==False:
            try:
                x=hdr[p]
                raise(PE.pyaOtherErrors(" Parameter "+str(p)+" already present in fits-header", where="Params::updateFitsHeader"))
                return
            except:
                pass
        hdr.update(p, model[p])
    if ic.check["pymc"]:
        from pymc.utils import hpd
        hdr.update('Conf',conf,"Error Confidence Level")
        for p in model.parameters():
            try:
                v_err=hpd(model.MCMC.trace(p)[:], 1.0-conf)
                p0,p1=str(p+'_e0'),str(p+'_e1')
                if len(p0) > 8:
                    raise(PE.pyaOtherErrors(" Cannot save Error for parameter "+str(p)+" because len(" + p0 + ")>8", where="Params::updateFitsHeader"))
                    return
                hdr.update(p0,v_err[0],"Lower confidence boundary")
                hdr.update(p1,v_err[1],"Upper confidence boundary")
            except:
                pass

    return hdr 

def restoreFromFits(filename,ext=1, verbose=1):
    if not ic.check["pyfits"]:
      raise(PE.PyARequiredImport("pyfits required to use fits file.", where="Params::updateFitsHeader"))
      return
    import pyfits
    if not os.path.isfile(filename):
        raise(PE.PyAValError("No such file: "+str(filename), where="restoreFromFits", solution="Give correct filename."))
        return
    ff=pyfits.open(filename)[ext]
    try:
        modelname=ff.header['PA_model']
    except:
        raise(PE.pyaOtherErrors("File ("+str(filename)+") contains no model.", where="restoreFromFits", solution="Use another file."))
        return
    if modelname=='convLine':
        import convLine as cl
        nLines=ff.header['model_p0']
        model=cl.ConvLine(nLines)
    elif modelname=='multiGauss1d':
        from PyAstronomy.funcFit import MultiGauss1d
        nLines=ff.header['model_p0']
        model=MultiGauss1d(nLines)
    for p in model.parameters():
        model[p]=ff.header[p]
    return model
    
        
     


# -*- coding: utf-8 -*-
from __future__ import print_function, division
from . import periodBase
import numpy
import scipy.special


class Fourier(periodBase.PeriodBase):
  r"""
    This class computes the Fast Fourier Transform (FFT) of the input data.
    
    This is actually a wrapper around numpy's FFT routines.
    The constructor takes the light
    curve, `lc` (*TimeSeries* instance), as input. The optional
    argument specifies the normalization of the Fourier Power.
    Currently, only the normalization according to [Leahy83]_
    is supported, which in the case of purely Poissonian
    noise results in a mean power of 2.
    
    Parameters
    ----------
    lc : TimeSeries instance
        The light curve to be analyzed.
    norm : optional, string
        Normalization method; currently, only default ("Leahy") is supported.
  """

  def __calcPeriodogram(self):

    dt=self.t[1:]-self.t[0:-1]
    if numpy.max(dt) > 1.1 * numpy.min(dt):
      print("timingAna::fourier() - Error: data must be equally spaced!")
      exit()

    power=numpy.abs(numpy.fft.fft(self.y))**2.
    if self.norm=="Leahy":
      power=power*2./numpy.sum(self.y)
    else:
      print("Normalization not yet implemented!")
      power=power*2./numpy.sum(self.y)
    freq = numpy.fft.fftfreq(numpy.size(self.y),numpy.mean(dt))
    indi = numpy.where(freq>0)
    self.freq = freq[indi]
    self.power = power[indi]
    self.M = len(self.freq)

  def __init__(self, lc, norm="Leahy"):
    self.freq = None
    self.power = None
    self.t = lc.time
    self.y = lc.flux
    self.norm = norm
    self.label = {'title': 'Fourier Transformation',\
                  'xlabel': 'Frequency',\
                  'ylabel': 'Fourier Power'}
    self.__calcPeriodogram()

  def Prob(self, Pn):
    r"""
    Returns the probability to obtain a power larger than the threshold, `Pn`.
    
    In the calculations, the noise is assumed to be Poisson-distributed.
    
    .. note::
        According to [vdK]_ the probability \
        to obtain a power larger than a given threshold from the noise is given by
        
        .. math::
          Prob(p>Pn) = Q(M \times W \times Pn, 2 \times M \times W)
        
        where :math:`Q(\chi^2, \nu)` is the cumulative :math:`\chi^2` distribution with :math:`\nu` d.o.f.

    Parameters
    ----------
    Pn : float
        Power threshold.
    
    Returns
    -------
    FAP : float
        The probability to obtain a power larger than the specified
        threshold from noise.
    """
    nu = 2
    Q = scipy.special.chdtrc(nu, Pn) #-- Integral from x to infinity of Chi-square pdf.
    return Q






from __future__ import print_function, division
from PyAstronomy.pyaC import pyaPermanent as pp
from PyAstronomy.pyaC import pyaErrors as PE
import PyAstronomy.pyaC as pyaC
import os
import gzip
import csv
import numpy as np
import six


class NasaExoplanetArchive(pp.PyAUpdateCycle):
    """
    Easy access to NASA's exoplanet archive.

    This class downloads a table of exoplanet data from
    the "NASA Exoplanet Archive"
    (http://exoplanetarchive.ipac.caltech.edu/index.html)
    and provides access to these data. By default, the
    data will be re-downloaded every seven days.

    The following data are provided:

    ===========   ===================================  ======
    Column        Description                          Unit
    -----------   -----------------------------------  ------
    pl_hostname   Name of host star
    pl_name       Name of the planet
    pl_letter     Planet letter (e.g., b, c, d, etc.)
    ra            Right ascension                      deg
    dec           Declination                          deg
    pl_orbper     Planetary orbital period             d
    pl_massj      Planetary mass                       MJ
    pl_radj       Planetary radius                     RJ
    pl_trandep    Central depth of transit             %
    pl_trandur    Transit duration                     d
    pl_tranmid    Transit midpoint                     BJD
    pl_orbsmax    Semi-major-axis                      AU
    pl_orbincl    Orbital inclination of planet        deg
    st_rad        Stellar radii                        Solar
    st_dist       Distance to star                     pc
    st_mass       Stellar mass                         Solar
    st_teff       Effective temperature of star        K
    st_vsin       Stellar vsin(i)                      km/s
    st_logg       Stellar surface gravity              cm/s**2
    sy_vmag       Stellar V-band brightness            mag
    sy_kmag       Stellar Ks-band brightness           mag
    ===========   ===================================  ======
    """

    def _downloadData(self):
        """
        Download data and store it to file in PyA's data directory.
        """
        urlRoot = "https://exoplanetarchive.ipac.caltech.edu/TAP/sync?query="
        select = "select+"
        v0s = [v[0] for v in six.itervalues(self._columns)]
        select += ",".join(v0s)
        select += "+from+ps"

        url = "".join((urlRoot, select))
        print("Requesting URL: ", url)
        self._fs.downloadToFile(
            url, self.dataFileName, clobber=True, verbose=False, openMethod=gzip.open
        )

    def __init__(self):
        self.data = None
        self.dataFileName = os.path.join("pyasl", "resBased", "NEXA.vo.gz")
        configFileName = os.path.join("pyasl", "resBased", "NEXA.cfg")
        pp.PyAUpdateCycle.__init__(self, configFileName, "NEXA")
        # Define columns to select
        # Column name, Description, Unit
        self._columns = {}
        self._columns[0] = ["hostname", "Name of host star", "", "U15"]
        self._columns[1] = ["pl_name", "Name of the planet", "", "U15"]
        self._columns[2] = [
            "pl_letter",
            "Planet letter (e.g., b, c, d, etc.)",
            "",
            "U2",
        ]
        self._columns[3] = ["ra", "Right ascension", "deg", float]
        self._columns[4] = ["dec", "Declination", "deg", float]
        self._columns[5] = ["pl_orbper", "Planetary orbital period", "d", float]
        self._columns[6] = ["pl_massj", "Planetary mass", "MJ", float]
        self._columns[7] = ["pl_radj", "Planetary radius", "RJ", float]
        self._columns[8] = ["pl_trandep", "Central depth of transit", "%", float]
        self._columns[9] = ["pl_trandur", "Transit duration", "d", float]
        self._columns[10] = ["pl_tranmid", "Transit midpoint", "BJD", float]
        self._columns[11] = ["pl_orbsmax", "Semi-major-axis", "AU", float]
        self._columns[12] = [
            "pl_orbincl",
            "Orbital inclination of planet",
            "deg",
            float,
        ]
        self._columns[13] = ["st_rad", "Stellar radii", "Solar", float]
        self._columns[14] = ["sy_dist", "Distance to star", "pc", float]
        self._columns[15] = ["st_mass", "Stellar mass", "Solar", float]
        self._columns[16] = ["st_teff", "Effective temperature of star", "K", float]
        self._columns[17] = ["st_logg", "Stellar surface gravity", "cm/s**2", float]
        self._columns[18] = ["sy_vmag", "Stellar V-band brightness", "mag", float]
        self._columns[19] = [
            "st_vsin",
            "Projected stellar rotation speed (vsini)",
            "km/s",
            float,
        ]
        self._columns[20] = ["sy_kmag", "Stellar Ks-band brightness", "mag", float]
        # Check whether data file exists
        self._fs = pp.PyAFS()

        if self.needsUpdate():
            # Data needs update
            print("Downloading exoplanet data from NASA exoplanet archive")
            self._update(self._downloadData)
            print("Saved data to file: ", self.dataFileName, " in data directory,")
            print("  which has been configured as: ", self._fs.dpath)
            print("By default, the data will be downloaded anew every 7 days.")
            print("You can use the `changeDownloadCycle` to change this behavior.")
        self._readData()

    def downloadData(self):
        """
        Trigger download of data.
        """
        self._update(self._downloadData)

    def _readData(self):
        """
        Read the data from local file into numpy recarray.
        """
        try:
            from astropy.io.votable import parse
        except ImportError as e:
            PE.PyAImportFailure(
                "Could not import astropy.io.votable to read VO table",
                solution="Please install astropy (e.g., via pip install astropy)",
            )
        if not self._fs.fileExists(self.dataFileName):
            print("Could not find data file. Initiating download.")
            self._update(self._downloadData)
        self.pstable = parse(
            self._fs.requestFile(self.dataFileName, "rb", gzip.open)
        ).get_first_table()

        dtype = [
            (self._columns[x][0], self._columns[x][3])
            for x in range(len(self._columns))
        ]

        self.data = np.rec.array(self.pstable.array)

    def availableColumns(self, verbose=True):
        """
        Shows a list of available data columns.

        Parameters
        ----------
        verbose : boolean, optional
            If True (default), prints information
            to screen.

        Returns
        -------
        columns : list of strings
            The names of the available data columns.
        """
        if self.data is None:
            return None
        if verbose:
            print("{0:12s}  {1:35s}  {2:5s}".format("Column", "Description", "Unit"))
            print("-" * 56)
            for v in six.itervalues(self._columns):
                print("{0:12s}  {1:35s}  {2:5s}".format(*v))

        return [self._columns[x][0] for x in range(len(self._columns))]

    def selectByPlanetName(self, planetName, caseSensitive=False):
        """
        Get entry by planet name.

        Parameters
        ----------
        planetName : string
            The name of the planet (includes planet letter,
            e.g., "corot-2 b"
        caseSensitive : boolean, optional
            If False (default), the search will be case-insensitive.

        Returns
        -------
        Data entry : dictionary
            A dictionary with a key for every data column holding
            the associated value from the data table.
        """
        if hasattr(self.data.pl_name[0], "decode"):
            pnames = [name.decode("utf8") for name in self.data.pl_name]
        else:
            pnames = self.data.pl_name
        r = pyaC.fuzzyMatch(
            planetName, pnames, caseSensitive=caseSensitive, raises=True
        )
        result = {}
        for c in six.itervalues(self._columns):
            result[c[0]] = self.data[c[0]][r["index"]]
        return result

    def getAllData(self):
        """
        Get all available data.

        Returns
        -------
        Data : numpy recarray
            All data stored in the table as a numpy recarray.
        """
        return self.data.copy()



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy as np
from .onedfit import OneDFit
from PyAstronomy.pyaC import pyaErrors as PE


class GaussFit1d(OneDFit):
    r"""
    A one-dimensional Gaussian

    The functional form is:

    .. math:: \frac{A}{\sqrt{2\pi\sigma^2}}e^{-(x-\mu)^2/(2\sigma^2)} + x \times lin + off

    Here, `lin` and `off` denote the linear and the offset term.

    *Fit parameters*:
     - `A` - Area of the Gaussian (formerly called Amplitude)
     - `mu` - Center of the Gaussian
     - `sig` - Standard deviation
     - `off` - Offset
     - `lin` - Linear term

    .. note::
        Other parameterizations using FWHM and 'height' of the curve can be used.

    Parameters
    ----------
    prm : optional tuple of two strings, {("A", "sig"), ("h", "sig"), ("A", "FWHM"), ("h", "FWHM")}
        Can be used to adapt the parameterization of the Gaussian. By default, the curve is
        parameterized by Area (A) and standard deviation (sig). Alternatively, also the height (h)
        of the curve (h = A/sqrt(2*pi*sig**2)) and the Full Width at Half Maximum (FWHM) can be
        used. The naming of the fit parameters is updated accordingly.
    """

    # Valid parameterizations
    _valid_prms = (("A", "sig"), ("h", "sig"), ("A", "FWHM"), ("h", "FWHM"))
    # FWHM -> STD
    _fwhm_to_sig = 2 * np.sqrt(2 * np.log(2))

    def __init__(self, prm=("A", "sig")):
        if not prm in GaussFit1d._valid_prms:
            raise (
                PE.PyAValError(
                    f"Invalid parameterization (f{str(prm)}).",
                    where="GaussFit1d",
                    solution="Choose any of: " +
                    ", ".join([str(x) for x in GaussFit1d._valid_prms]),
                )
            )

        self._prm = prm
        OneDFit.__init__(self, ["mu", "off", "lin"] + list(prm))
        self.setRootName("Gaussian")

    def _check_zero_width(self):
        r"""
        Check if width (sig/FWHM) is zero. Raise exception if so.
        """
        if self[self._prm[1]] == 0.0:
            raise (
                PE.PyAValError(
                    "Width of Gaussian must be larger than zero.",
                    solution=f"Change width ('{self._prm[1]}').",
                )
            )

    def evaluate_h_sig(self, x, h, sig):
        r"""
        Evaluates the model for given height and standard deviation (sig)-

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.

        Returns
        -------
        model : array
        """

        y = (
            h * np.exp(-((self["mu"] - x) ** 2) / (2.0 * sig**2))
            + self["off"]
            + (self["lin"] * x)
        )
        return y

    def evaluate(self, x):
        self._check_zero_width()
        if self._prm == ("h", "sig"):
            return self.evaluate_h_sig(x, self["h"], self["sig"])
        elif self._prm == ("A", "sig"):
            return self.evaluate_h_sig(
                x, self["A"] / np.sqrt(2 * np.pi *
                                       self["sig"] ** 2), self["sig"]
            )
        elif self._prm == ("h", "FWHM"):
            return self.evaluate_h_sig(
                x, self["h"], self["FWHM"] / GaussFit1d._fwhm_to_sig
            )
        elif self._prm == ("A", "FWHM"):
            return self.evaluate_h_sig(
                x,
                self["A"] / np.sqrt(2 * np.pi *
                                    (self["FWHM"]/self._fwhm_to_sig) ** 2),
                self["FWHM"] / GaussFit1d._fwhm_to_sig,
            )


class MultiGauss1d(OneDFit):
    r"""
    A multicomponent Gaussian with a single linear continuum component.

    The parameters are the same as for the *GaussFit1d*,
    except that all receive a number specifying the Gaussian
    component to which they belong. Therefore, they are, e.g.,
    named `A1`, `mu2`, and so on, only `off` and `lin`
    remain unnumbered.

    Parameters
    ----------
    n : int
        The number if Gaussian components.
    """

    def __init__(self, n):
        self.n = n
        params = ["off", "lin"]
        for i in range(n):
            p = str(i + 1)
            params.extend(["A" + p, "mu" + p, "sig" + p])
        OneDFit.__init__(self, params)
        self.setRootName("MultiGauss")

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        y = self["off"] + self["lin"] * x
        for i in range(self.n):
            p = str(i + 1)
            y += (
                self["A" + p]
                / np.sqrt(2.0 * np.pi * self["sig" + p] ** 2)
                * np.exp(-((self["mu" + p] - x) ** 2) / (2.0 * self["sig" + p] ** 2))
            )
        return y

    def evalComponent(self, x, p):
        r"""
        Evaluate the model considering only a single component.

        Parameters
        ----------
        x : array
            The abscissa.
        p : int
            Component number (starts with one).

        Returns
        -------
        single component model : array
            The model considering a single component.
        """
        if p > 0 and p <= self.n:
            p = str(p)
            y = self["off"] + self["lin"] * x
            y += (
                self["A" + p]
                / np.sqrt(2.0 * np.pi * self["sig" + p] ** 2)
                * np.exp(-((self["mu" + p] - x) ** 2) / (2.0 * self["sig" + p] ** 2))
            )
            return y
        else:
            raise (
                PE.PyAValError(
                    "No such component (no. " + str(p) + ")",
                    where="MultiGauss1d::evalComponent",
                    solution="Use value between 1 and " + str(self.n),
                )
            )



from __future__ import print_function, division
from numpy import pi, sqrt, exp, meshgrid, shape, ones
from .onedfit import OneDFit
from PyAstronomy.pyaC import pyaErrors as PE


class GaussFit2d(OneDFit):
    r"""
    Implements a two dimensional Gaussian.

    Expects a coordinate array to evaluate model.

    The functional form is:

    .. math:: \frac{A}{2\pi\sigma_x\sigma_y\sqrt{1-\rho^2}}
              exp\left(-\frac{1}{2(1-\rho^2)}\left( \frac{(x-\mu_x)^2}{\sigma_x^2} +
              \frac{(y-\mu_y)^2}{\sigma_y^2} -
              \frac{2\rho(x-\mu_x)(y-\mu_y)}{\sigma_x\sigma_y}
              \right)\right)

    Here, `lin` and `off` denote the linear and the offset term.

    *Fit parameters*:
     - `A` - Amplitude (the area of the Gaussian)
     - `mux` - Center of the Gaussian (x-axis)
     - `muy` - Center of the Gaussian (y-axis)
     - `sigx` - Standard deviation (x-axis)
     - `sigy` - Standard deviation (y-axis)
     - `rho` - Correlation
    """

    def __init__(self):
        OneDFit.__init__(self, ["A", "mux", "muy", "sigx", "sigy", "rho"])
        self.setRootName("Gaussian2d")

    def evaluate(self, co):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        co : array
             Specifies the points at which to evaluate the model.
        """
        if (self["sigx"] <= 0.0) or (self["sigy"] <= 0.0):
            raise (PE.PyAValError("Width(s) of Gaussian must be larger than zero.",
                                  solution="Change width ('sigx/y')."))
        if self["rho"] > 1.0:
            raise (PE.PyAValError("The correlation coefficient must be 0 <= rho <= 1.",
                                  solution="Change width ('sigx/y')."))
        result = self["A"]/(2.*pi*self["sigx"]*self["sigy"]*sqrt(1.-self["rho"]**2)) * \
            exp(((co[::, ::, 0]-self["mux"])**2/self["sigx"]**2 + (co[::, ::, 1]-self["muy"])**2/self["sigy"]**2 -
                 2.*self["rho"]*(co[::, ::, 0]-self["mux"])*(co[::, ::, 1]-self["muy"])/(self["sigx"]*self["sigy"])) /
                (-2.*(1.-self["rho"]**2)))
        return result


class MultiGauss2d(OneDFit):
    r"""
    Implements a multicomponent, two dimensional Gaussian.

    Expects a coordinate array to evaluate model.

    The functional form is:

    .. math:: \frac{A}{2\pi\sigma_x\sigma_y\sqrt{1-\rho^2}}
              exp\left(-\frac{1}{2(1-\rho^2)}\left( \frac{(x-\mu_x)^2}{\sigma_x^2} +
              \frac{(y-\mu_y)^2}{\sigma_y^2} -
              \frac{2\rho(x-\mu_x)(y-\mu_y)}{\sigma_x\sigma_y}
              \right)\right)

    Parameters
    ----------
    n : int
        The number of Gaussian components.

    *Fit parameters*:
     - `A` - Amplitude (the area of the Gaussian)
     - `mux` - Center of the Gaussian (x-axis)
     - `muy` - Center of the Gaussian (y-axis)
     - `sigx` - Standard deviation (x-axis)
     - `sigy` - Standard deviation (y-axis)
     - `rho` - Correlation
     - `off` - Constant offset
    """

    def __init__(self, n):
        params = ["off"]
        for i in range(n):
            p = str(i+1)
            params.extend(["A"+p, "mux"+p, "muy"+p,
                          "sigx"+p, "sigy"+p, "rho"+p])
        self.n = n
        OneDFit.__init__(self, params)
        self.setRootName("MultiGauss2d")

    def evaluate(self, co):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        co : array
             Specifies the points at which to evaluate the model.
        """

        result = ones((shape(co)[0], shape(co)[1])) * self["off"]

        for i in range(self.n):
            p = str(i+1)
            if (self["sigx"+p] <= 0.0) or (self["sigy"+p] <= 0.0):
                raise (PE.PyAValError("Width(s) of Gaussian must be larger than zero.",
                                      solution="Change width ('sigx/y')."))
            if self["rho"+p] > 1.0:
                raise (PE.PyAValError("The correlation coefficient must be 0 <= rho <= 1.",
                                      solution="Change width ('sigx/y')."))

            result += self["A"+p]/(2.*pi*self["sigx"+p]*self["sigy"+p]*sqrt(1.-self["rho"+p]**2)) * \
                exp(((co[::, ::, 0]-self["mux"+p])**2/self["sigx"+p]**2 + (co[::, ::, 1]-self["muy"+p])**2/self["sigy"+p]**2 -
                     2.*self["rho"+p]*(co[::, ::, 0]-self["mux"+p])*(co[::, ::, 1]-self["muy"+p])/(self["sigx"+p]*self["sigy"+p])) /
                    (-2.*(1.-self["rho"+p]**2)))
        return result


class GaussFit2dTuple(OneDFit):
    r"""
    Implements a two dimensional Gaussian.

    Expects a tuple or list of coordinate axes to evaluate the model.

    The functional form is:

    .. math:: \frac{A}{2\pi\sigma_x\sigma_y\sqrt{1-\rho^2}}
              exp\left(-\frac{1}{2(1-\rho^2)}\left( \frac{(x-\mu_x)^2}{\sigma_x^2} +
              \frac{(y-\mu_y)^2}{\sigma_y^2} -
              \frac{2\rho(x-\mu_x)(y-\mu_y)}{\sigma_x\sigma_y}
              \right)\right)

    Here, `lin` and `off` denote the linear and the offset term.

    *Fit parameters*:
     - `A` - Amplitude (the area of the Gaussian)
     - `mux` - Center of the Gaussian (x-axis)
     - `muy` - Center of the Gaussian (y-axis)
     - `sigx` - Standard deviation (x-axis)
     - `sigy` - Standard deviation (y-axis)
     - `rho` - Correlation
    """

    def __init__(self):
        OneDFit.__init__(self, ["A", "mux", "muy", "sigx", "sigy", "rho"])
        self.setRootName("Gaussian2d")

    def evaluate(self, co):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        co : array
             Specifies the points at which to evaluate the model.
        """
        if (self["sigx"] <= 0.0) or (self["sigy"] <= 0.0):
            raise (PE.PyAValError("Width(s) of Gaussian must be larger than zero.",
                                  solution="Change width ('sigx/y')."))
        if self["rho"] > 1.0:
            raise (PE.PyAValError("The correlation coefficient must be 0 <= rho <= 1.",
                                  solution="Change width ('sigx/y')."))
        xx, yy = meshgrid(co[0], co[1])
        result = self["A"]/(2.*pi*self["sigx"]*self["sigy"]*sqrt(1.-self["rho"]**2)) * \
            exp(((xx-self["mux"])**2/self["sigx"]**2 + (yy-self["muy"])**2/self["sigy"]**2 -
                 2.*self["rho"]*(xx-self["mux"])*(yy-self["muy"])/(self["sigx"]*self["sigy"])) /
                (-2.*(1.-self["rho"]**2)))
        return result



from PyAstronomy.pyaC import _ic
from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np


def ibtrapz(x, y, x0, x1, iaout=False, jne=False):
    """
    Use the trapezoid rule to integrate and interpolate boundary values.

    Can be used for integration on tabled data, where the integration boundaries
    are not located on the grid. The values to use at the boundaries are
    determined using linear interpolation.

    Parameters
    ----------
    x,y : arrays
        The data.
    x0, x1 : float
        The integration boundaries.
    iaout : boolean, optional
        If True, the arrays used internally for
        integration are also returned. The default
        is False.
    jne : boolean, optional
        If True, a leave-one-out error estimate
        is produced for the value of the integral.
        In that case, the actual integral, the error
        estimate (standard deviation), and the individual
        jacknife estimates are returned.

    Returns
    -------
    Integral : float
        The value of the resulting integral.
    xi, yi : arrays, optional
        Internally used arrays for integration including
        the values derived at the boundaries. Only returned
        if `iaout` is set to True.
    """
    if not _ic.check["scipy"]:
        raise (
            PE.PyARequiredImport(
                "scipy could not be imported.",
                where="ibtrapz",
                solution="Please install scipy to use ibtrapz.",
            )
        )
    import scipy.interpolate as sci
    import scipy.integrate as scinteg

    if x1 < x0:
        raise (
            PE.PyAValError(
                "x1 must be larger than x0.",
                where="ibtrapz",
                solution="Exchange boundaries.",
            )
        )
    if np.any(x[1:] - x[0:-1] < 0.0):
        raise (
            PE.PyAValError(
                "x axis must be in ascending order.",
                where="ibtrapz",
                solution="Sort values.",
            )
        )
    if (x0 < np.min(x)) or (x1 > np.max(x)):
        raise (
            PE.PyAValError(
                "One or both integration boundaries are beyond the range of data given.",
                where="ibtrapz",
                solution="Give more data or adjust integration boundaries.",
            )
        )
    if jne & iaout:
        raise (
            PE.PyAParameterConflict(
                "jne and iaout cannot be used simultaneously.", where="ibtrapz"
            )
        )

    # Interpolate at integration boundaries
    fi = sci.interp1d(x, y)
    y0, y1 = fi(x0), fi(x1)
    # Search data within boundaries
    indi = np.where((x > x0) & (x < x1))[0]
    # Construct arrays to be used in integration
    xi = np.zeros(len(indi) + 2)
    yi = np.zeros(len(indi) + 2)
    xi[1:-1] = x[indi]
    yi[1:-1] = y[indi]
    xi[0] = x0
    xi[-1] = x1
    yi[0] = y0
    yi[-1] = y1

    if jne:
        # Produce a leave-one-out jackknife error
        if len(xi) < 3:
            # This is technically impossible
            return scinteg.trapezoid(yi, xi), None, None
        m = np.ones_like(xi, dtype=bool)
        estis = []
        for i in range(1, len(m) - 1):
            m[i] = False
            estis.append(scinteg.trapezoid(yi[m], xi[m]))
            m[i] = True
        estis = np.array(estis)
        me = np.mean(estis)
        var = (len(estis) - 1) / len(estis) * np.sum((estis - me) ** 2)
        return scinteg.trapezoid(yi, xi), np.sqrt(var), estis

    if iaout:
        return scinteg.trapezoid(yi, xi), xi, yi
    return scinteg.trapezoid(yi, xi)
